@@ -1,7 +1,7 @@
 /*
  *  UFTP - UDP based FTP with multicast
  *
- *  Copyright (C) 2001-2019   Dennis A. Bush, Jr.   bush@tcnj.edu
+ *  Copyright (C) 2001-2020   Dennis A. Bush, Jr.   bush@tcnj.edu
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -313,7 +313,7 @@ void run_postreceive_multi(struct group_list_t *group, char *const *files,
             is_cmd = 0;
         }
         for (too_long = 0, i = 0; i < count + 3; i++) {
-            int size = 0x8000 - strlen(cmdline);
+            int size = 0x8000 - (int)strlen(cmdline);
             if (size <= (int)strlen(params[i]) + 4) {
                 too_long = 1;
                 break;
@@ -360,7 +360,7 @@ void run_postreceive_multi(struct group_list_t *group, char *const *files,
             close(2);
             execv(postreceive, params);
             gsyserror(group, "exec failed");
-            exit(1);
+            _exit(1);
         }
     }
 #endif
@@ -421,12 +421,17 @@ void file_cleanup(struct group_list_t *group, int abort_session)
     if (abort_session || (group->file_id == 0)) {
         if (!addr_blank(&group->multi) && !other_mcast_users(group) &&
                 group->multi_join) {
-            if (server_count > 0) {
+            if (use_ssm) {
                 multicast_leave(listener, group->group_id, &group->multi,
-                        m_interface, interface_count, server_keys,server_count);
-                if (has_proxy) {
+                        m_interface, interface_count, server_list,server_count);
+                if (proxy_count) {
                     multicast_leave(listener, group->group_id, &group->multi,
-                            m_interface, interface_count, &proxy_info, 1);
+                                    m_interface, interface_count,
+                                    proxy_list, proxy_count);
+                }
+                if (has_v4_proxy) {
+                    multicast_leave(listener, group->group_id, &group->multi,
+                            m_interface, interface_count, &proxy4_info, 1);
                 }
             } else {
                 multicast_leave(listener, group->group_id, &group->multi,
@@ -434,25 +439,42 @@ void file_cleanup(struct group_list_t *group, int abort_session)
             }
         }
         if (group->server_pubkey.key) {
-            if (group->keyextype == KEYEX_ECDH_ECDSA) {
+            if (group->server_keytype == KEYBLOB_EC) {
                 free_EC_key(group->server_pubkey.ec);
             } else {
                 free_RSA_key(group->server_pubkey.rsa);
             }
         }
-        if (group->server_dhkey.key) {
-            free_EC_key(group->server_dhkey.ec);
-            free_EC_key(group->client_dhkey.ec);
+        if (group->proxy_pubkey.key) {
+            if (group->proxy_keytype == KEYBLOB_EC) {
+                free_EC_key(group->proxy_pubkey.ec);
+            } else {
+                free_RSA_key(group->proxy_pubkey.rsa);
+            }
         }
+        free_EC_key(group->server_dhkey.ec);
+        free_EC_key(group->client_dhkey.ec);
+        free_EC_key(group->proxy_dhkey.ec);
+        free(group->s_context);
+        free(group->p_context);
+        free(group->c_context1);
+        free(group->c_context2);
         if (group->restartinfo &&
                 (strcmp(group->restartinfo->name, ""))) {
             // We have unused restart info from the last run.
             // Chalk this up as a loss and delete the data file
             char filepath[MAXPATHNAME];
-            snprintf(filepath, sizeof(filepath), "%s%c_group_%08X%c%s", tempdir,
-                     PATH_SEP, group->group_id, PATH_SEP,
-                     group->restartinfo->name);
-            unlink(filepath);
+            int len;
+            len = snprintf(filepath, sizeof(filepath), "%s%c_group_%08X%c%s",
+                           tempdir, PATH_SEP, group->group_id, PATH_SEP,
+                           group->restartinfo->name);
+            if ((len >= sizeof(filepath)) || (len == -1)) {
+                glog0(group,"Max pathname length exceeded: %s%c_group_%08X%c%s",
+                            tempdir, PATH_SEP, group->group_id, PATH_SEP,
+                            group->restartinfo->name);
+            } else {
+                unlink(filepath);
+            }
         }
         if (abort_session) {
             write_restart_file(group);
@@ -501,8 +523,8 @@ int flush_disk_cache(struct group_list_t *group)
         gsyserror(group, "lseek failed for file");
     }
     if (seek_rval != offset) {
-        glog2(group, "offset is %s", printll(seek_rval));
-        glog2(group, "  should be %s", printll(offset));
+        glog2(group, "offset is " F_u64, seek_rval);
+        glog2(group, "  should be " F_u64, offset);
         if ((seek_rval = lseek_func(group->fileinfo.fd, offset,
                                     SEEK_SET)) == -1) {
             gsyserror(group, "lseek failed for file");
@@ -569,8 +591,9 @@ void set_timeout(struct group_list_t *group, int rescale)
     }
     group->timeout_time = group->start_timeout_time;
     switch (group->phase) {
+    case PHASE_PREREGISTER:
     case PHASE_REGISTERED:
-        add_timeval_d(&group->timeout_time, 4 * group->grtt);
+        add_timeval_d(&group->timeout_time, 5 * group->grtt);
         break;
     case PHASE_RECEIVING:
     case PHASE_MIDGROUP:
@@ -609,13 +632,12 @@ void send_abort(struct group_list_t *group, const char *message)
 
     payloadlen = sizeof(struct abort_h);
     if ((group->phase != PHASE_REGISTERED) &&
+            (group->phase != PHASE_PREREGISTER) &&
             (group->keytype != KEY_NONE)) {
         encrypted = NULL;
         if (!encrypt_and_sign(buf, &encrypted, payloadlen, &enclen,
-                group->keytype, group->groupkey, group->groupsalt,&group->ivctr,
-                group->ivlen, group->hashtype, group->grouphmackey,
-                group->hmaclen, group->sigtype, group->keyextype,
-                group->client_privkey, group->client_privkeylen)) {
+                group->keytype, group->c_app_key, group->c_app_iv,
+                &group->ivctr, group->ivlen)) {
             glog0(group, "Error encrypting ABORT");
             free(buf);
             return;
@@ -691,14 +713,14 @@ void send_key_req()
 
     header = (struct uftp_h *)packet;
     keyreq = (struct key_req_h *)(packet + sizeof(struct uftp_h));
-    header->version = UFTP_VER_NUM;
+    header->version = UFTP4_VER_NUM;
     header->func = KEY_REQ;
     header->src_id = uid;
     keyreq->func = KEY_REQ;
     keyreq->hlen = sizeof(struct key_req_h) / 4;
 
     meslen = sizeof(struct uftp_h) + sizeof(struct key_req_h);
-    proxyaddr = proxy_info.addr;
+    proxyaddr = proxy4_info.addr;
     if (nb_sendto(listener, packet, meslen, 0, 
                   (struct sockaddr *)&proxyaddr,
                   family_len(proxyaddr)) == SOCKET_ERROR) {
@@ -719,25 +741,25 @@ void send_key_req()
 }
 
 /**
- * Process a PROXY_KEY message
+ * Process a V4 PROXY_KEY message
  */
-void handle_proxy_key(const union sockaddr_u *src,
-                      unsigned char *message, unsigned meslen)
+void handle_v4_proxy_key(const union sockaddr_u *src,
+                         unsigned char *message, unsigned meslen)
 {
-    struct proxy_key_h *proxykey;
+    struct proxy_key_4_h *proxykey;
     unsigned char *keyblob, *dhblob, *sig;
     unsigned char fingerprint[HMAC_LEN];
     unsigned int fplen, keylen, dhlen, siglen;
     char addrname[INET6_ADDRSTRLEN];
     int rval;
 
-    proxykey = (struct proxy_key_h *)message;
+    proxykey = (struct proxy_key_4_h *)message;
 
     if (meslen < (proxykey->hlen * 4U) ||
-            ((proxykey->hlen * 4U) < sizeof(struct proxy_key_h) +
+            ((proxykey->hlen * 4U) < sizeof(struct proxy_key_4_h) +
                 ntohs(proxykey->bloblen) + ntohs(proxykey->dhlen) +
                 ntohs(proxykey->siglen))) {
-        log2(0, 0, 0, "Rejecting PROXY_KEY: invalid message size");
+        log2(0, 0, 0, "Rejecting PROXY_KEY_4: invalid message size");
         return;
     }
 
@@ -746,72 +768,80 @@ void handle_proxy_key(const union sockaddr_u *src,
             NULL, 0, NI_NUMERICHOST)) != 0) {
         log1(0, 0, 0, "getnameinfo failed: %s", gai_strerror(rval));
     }
-    log2(0, 0, 0, "Received PROXY_KEY from %s", addrname);
+    log2(0, 0, 0, "Received PROXY_KEY_4 from %s", addrname);
 
-    if (!has_proxy) {
+    if (!has_v4_proxy) {
         log2(0, 0, 0, "No reply proxy specified");
         return;
     }
-    if (!addr_equal(&proxy_info.addr, src)) {
-        log2(0, 0, 0, "PROXY_KEY not from specified reply proxy");
+    if (!addr_equal(&proxy4_info.addr, src)) {
+        log2(0, 0, 0, "PROXY_KEY_4 not from specified reply proxy");
         return;
     }
 
-    keyblob = (unsigned char *)proxykey + sizeof(struct proxy_key_h);
+    keyblob = (unsigned char *)proxykey + sizeof(struct proxy_key_4_h);
     keylen = ntohs(proxykey->bloblen);
     dhblob = keyblob + keylen;
     dhlen = ntohs(proxykey->dhlen);
     sig = dhblob + dhlen;
     siglen = ntohs(proxykey->siglen);
+    if (!dhlen || !keylen) {
+        log2(0, 0, 0, "Missing key in PROXY_KEY_4");
+        return;
+    }
 
-    if (keyblob[0] == KEYBLOB_RSA) {
-        if (!import_RSA_key(&proxy_pubkey.rsa, keyblob, keylen)) {
-            log0(0, 0, 0, "Failed to import public key from PROXY_KEY");
+    if (!import_EC_key(&proxy4_dhkey.ec, dhblob, dhlen, 1)) {
+        log0(0, 0, 0, "Failed to import ECDH public key from PROXY_KEY_4");
+        return;
+    } 
+    proxy4_pubkeytype = keyblob[0];
+    if (proxy4_pubkeytype == KEYBLOB_RSA) {
+        if (!import_RSA_key(&proxy4_pubkey.rsa, keyblob, keylen)) {
+            log0(0, 0, 0, "Failed to import public key from PROXY_KEY_4");
+            free_EC_key(proxy4_dhkey.ec);
             return;
         } 
-        if (proxy_info.has_fingerprint) {
+        if (proxy4_info.has_fingerprint) {
             hash(HASH_SHA1, keyblob, keylen, fingerprint, &fplen);
-            if (memcmp(proxy_info.fingerprint, fingerprint, fplen)) {
-                log1(0, 0, 0, "Failed to verify PROXY_KEY fingerprint");
-                free_RSA_key(proxy_pubkey.rsa);
+            if (memcmp(proxy4_info.fingerprint, fingerprint, fplen)) {
+                log1(0, 0, 0, "Failed to verify PROXY_KEY_4 fingerprint");
+                free_EC_key(proxy4_dhkey.ec);
+                free_RSA_key(proxy4_pubkey.rsa);
                 return;
             }
         }
-        if (!verify_RSA_sig(proxy_pubkey.rsa, HASH_SHA1,
+        if (!verify_RSA_sig(proxy4_pubkey.rsa, HASH_SHA1,
                             (unsigned char *)&proxykey->nonce,
                             sizeof(proxykey->nonce), sig, siglen)) {
-            log1(0, 0, 0, "Failed to verify PROXY_KEY signature");
-            free_RSA_key(proxy_pubkey.rsa);
+            log1(0, 0, 0, "Failed to verify PROXY_KEY_4 signature");
+            free_EC_key(proxy4_dhkey.ec);
+            free_RSA_key(proxy4_pubkey.rsa);
             return;
         }
     } else {
-        if (!import_EC_key(&proxy_pubkey.ec, keyblob, keylen, 0)) {
-            log0(0, 0, 0, "Failed to import public key from PROXY_KEY");
+        if (!import_EC_key(&proxy4_pubkey.ec, keyblob, keylen, 0)) {
+            log0(0, 0, 0, "Failed to import public key from PROXY_KEY_4");
+            free_EC_key(proxy4_dhkey.ec);
             return;
         } 
-        if (proxy_info.has_fingerprint) {
+        if (proxy4_info.has_fingerprint) {
             hash(HASH_SHA1, keyblob, keylen, fingerprint, &fplen);
-            if (memcmp(proxy_info.fingerprint, fingerprint, fplen)) {
-                log1(0, 0, 0, "Failed to verify PROXY_KEY fingerprint");
-                free_RSA_key(proxy_pubkey.rsa);
+            if (memcmp(proxy4_info.fingerprint, fingerprint, fplen)) {
+                log1(0, 0, 0, "Failed to verify PROXY_KEY_4 fingerprint");
+                free_EC_key(proxy4_dhkey.ec);
+                free_EC_key(proxy4_pubkey.ec);
                 return;
             }
         }
-        if (!verify_ECDSA_sig(proxy_pubkey.ec, HASH_SHA1,
+        if (!verify_ECDSA_sig(proxy4_pubkey.ec, HASH_SHA1,
                               (unsigned char *)&proxykey->nonce,
                               sizeof(proxykey->nonce), sig, siglen)) {
-            log1(0, 0, 0, "Failed to verify PROXY_KEY signature");
-            free_RSA_key(proxy_pubkey.rsa);
+            log1(0, 0, 0, "Failed to verify PROXY_KEY_4 signature");
+            free_EC_key(proxy4_dhkey.ec);
+            free_EC_key(proxy4_pubkey.ec);
             return;
         }
     }
-    proxy_pubkeytype = keyblob[0];
-    if (dhlen) {
-        if (!import_EC_key(&proxy_dhkey.ec, dhblob, dhlen, 1)) {
-            log0(0, 0, 0, "Failed to import ECDH public key from PROXY_KEY");
-            return;
-        } 
-    }
 }
 
 /**
