@@ -1,7 +1,7 @@
 /*
  *  UFTP - UDP based FTP with multicast
  *
- *  Copyright (C) 2001-2015   Dennis A. Bush, Jr.   bush@tcnj.edu
+ *  Copyright (C) 2001-2020   Dennis A. Bush, Jr.   bush@tcnj.edu
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -79,6 +79,7 @@ int read_fileinfo(struct group_list_t *group, const unsigned char *message,
     const uint32_t *addrlist;
     int listlen, maxsecsize;
     const char *name, *flink, *p;
+    int64_t tstamp_usec;
 
     fileinfo = (const struct fileinfo_h *)message;
     addrlist = (const uint32_t *)(message + (fileinfo->hlen * 4));
@@ -148,9 +149,16 @@ int read_fileinfo(struct group_list_t *group, const unsigned char *message,
         group->fileinfo.big_sections = 0;
     }
 
-    group->fileinfo.tstamp = ntohl(fileinfo->ftstamp);
-    group->last_server_ts.tv_sec = ntohl(fileinfo->tstamp_sec);
-    group->last_server_ts.tv_usec = ntohl(fileinfo->tstamp_usec);
+    group->fileinfo.tstamp = (int64_t)ntohl(fileinfo->ftstamp_hi);
+    group->fileinfo.tstamp |= ntohl(fileinfo->ftstamp_lo);
+    if (group->version == UFTP4_VER_NUM) {
+        group->last_server_ts.tv_sec = ntohl(fileinfo->tstamp_hi);
+        group->last_server_ts.tv_usec = ntohl(fileinfo->tstamp_lo);
+    } else {
+        tstamp_usec = (int64_t)ntohl(fileinfo->tstamp_hi) << 32;
+        tstamp_usec |= ntohl(fileinfo->tstamp_lo);
+        group->last_server_ts = usec_to_tv(tstamp_usec);
+    }
     group->last_server_rx_ts = rxtime;
     group->fileinfo.fd = -1;
 
@@ -485,8 +493,8 @@ void handle_fileinfo(struct group_list_t *group, const unsigned char *message,
     glog2(group, "Name of file to receive: %s", group->fileinfo.name);
     switch (group->fileinfo.ftype) {
     case FTYPE_REG:
-        glog2(group, "Bytes: %s, Blocks: %d, Sections: %d",
-                     printll(group->fileinfo.size),
+        glog2(group, "Bytes: " F_i64 ", Blocks: %d, Sections: %d",
+                     group->fileinfo.size,
                      group->fileinfo.blocks, group->fileinfo.sections);
         glog3(group, "small section size: %d, "
                      "big section size: %d, # big sections: %d",
