@@ -1,7 +1,7 @@
 /*
  *  UFTP - UDP based FTP with multicast
  *
- *  Copyright (C) 2001-2019   Dennis A. Bush, Jr.   bush@tcnj.edu
+ *  Copyright (C) 2001-2020   Dennis A. Bush, Jr.   bush@tcnj.edu
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -73,17 +73,20 @@ void cleanup(void)
     for (i = 0; i < MAXLIST; i++) {
         if (group_list[i].group_id != 0) {
             send_abort(&group_list[i], "Client shutting down");
-            file_cleanup(&group_list[i], 1);
         }
     }
     if (!parent) {
         for (i = 0; i < pub_multi_count; i++) {
-            if (server_count > 0) {
+            if (use_ssm) {
                 multicast_leave(listener, 0, &pub_multi[i], m_interface,
-                                interface_count, server_keys, server_count);
-                if (has_proxy) {
+                                interface_count, server_list, server_count);
+                if (proxy_count) {
                     multicast_leave(listener, 0, &pub_multi[i], m_interface,
-                                    interface_count, &proxy_info, 1);
+                                    interface_count, proxy_list, proxy_count);
+                }
+                if (has_v4_proxy) {
+                    multicast_leave(listener, 0, &pub_multi[i], m_interface,
+                                    interface_count, &proxy4_info, 1);
                 }
             } else {
                 multicast_leave(listener, 0, &pub_multi[i], m_interface,
@@ -93,13 +96,17 @@ void cleanup(void)
     }
     closesocket(listener);
 
-    for (i = 0; i < key_count; i++) {
-        if (privkey_type[i] == KEYBLOB_RSA) {
-            free_RSA_key(privkey[i].rsa);
-        } else {
-            free_EC_key(privkey[i].ec);
-        }
+    if (privkey_type == KEYBLOB_RSA) {
+        free_RSA_key(privkey.rsa);
+    } else {
+        free_EC_key(privkey.ec);
     }
+    if (proxy4_pubkeytype == KEYBLOB_RSA) {
+        free_RSA_key(proxy4_pubkey.rsa);
+    } else if (proxy4_pubkeytype == KEYBLOB_EC) {
+        free_EC_key(proxy4_pubkey.ec);
+    }
+    free_EC_key(proxy4_dhkey.ec);
     crypto_cleanup();
 
 #ifdef WINDOWS
@@ -114,7 +121,7 @@ void cleanup(void)
 void gotsig(int sig)
 {
     log0(0, 0, 0, "Exiting on signal %d", sig);
-    exit(ERR_INTERRUPTED);
+    user_abort = 1;
 }
 
 #ifdef WINDOWS
@@ -143,7 +150,8 @@ BOOL WINAPI winsig(DWORD event)
         log0(0, 0, 0, "GOT unknown event %d", event);
         break;
     }
-    exit(ERR_INTERRUPTED);
+    user_abort = 1;
+    return TRUE;
 }
 #endif
 
@@ -260,63 +268,48 @@ void daemonize(void)
 void key_init(void)
 {
 #ifndef NO_ENCRYPTION
-    char *keyname;
-    int size, i;
+    int size;
     uint8_t curve;
 
     crypto_init(sys_keys);
 
-    if ((keyfile_count == 0) && (keyinfo_count == 0)) {
-        privkey[0].rsa = gen_RSA_key(0, RSA_EXP, NULL);
-        if (!privkey[0].key) {
+    if (!strcmp(keyfile, "") && !strcmp(keyinfo, "")) {
+        privkey.ec = gen_EC_key(DEF_CURVE, 0, NULL);
+        if (!privkey.key) {
             exit(ERR_CRYPTO);
         }
-        privkey_type[0] = KEYBLOB_RSA;
-        key_count = 1;
-    } else if (keyinfo_count != 0) {
-        key_count = 0;
-        for (i = 0; i < keyinfo_count; i++) {
-            if (keyfile_count <= i) {
-                keyname = NULL;
-            } else {
-                keyname = keyfile[i];
-            }
-            if (!strncmp(keyinfo[i], "ec:", 3)) {
-                curve = get_curve(&keyinfo[i][3]);
-                if (curve == 0) {
-                    log0(0, 0, 0, "Invalid EC curve: %s", &keyinfo[i][3]);
-                    exit(ERR_PARAM);
-                }
-                privkey[key_count].ec = gen_EC_key(curve, 0, keyname);
-                privkey_type[key_count] = KEYBLOB_EC;
-                if (!privkey[key_count].key) {
-                    exit(ERR_CRYPTO);
-                }
-            } else if (!strncmp(keyinfo[i], "rsa:", 4)) {
-                size = atoi(&keyinfo[i][4]);
-                if ((size < 512) || (size > 2048)) {
-                    log0(0, 0, 0, "Invalid RSA key size: %s", &keyinfo[i][4]);
-                    exit(ERR_PARAM);
-                }
-                privkey[key_count].rsa = gen_RSA_key(size, RSA_EXP, keyname);
-                privkey_type[key_count] = KEYBLOB_RSA;
-                if (!privkey[key_count].key) {
-                    exit(ERR_CRYPTO);
-                }
-            } else {
-                log0(0, 0, 0, "Invalid keyinfo entry: %s", keyinfo[i]);
+        privkey_type = KEYBLOB_EC;
+    } else if (strcmp(keyinfo, "")) {
+        if (!strncmp(keyinfo, "ec:", 3)) {
+            curve = get_curve(&keyinfo[3]);
+            if (curve == 0) {
+                log0(0, 0, 0, "Invalid EC curve: %s", &keyinfo[3]);
                 exit(ERR_PARAM);
             }
-            key_count++;
-        }
-    } else {
-        for (i = 0; i < keyfile_count; i++) {
-            privkey[key_count] =
-                    read_private_key(keyfile[i], &privkey_type[key_count]);
-            if (privkey_type[key_count] == 0) {
+            privkey.ec = gen_EC_key(curve, 0, keyfile);
+            privkey_type = KEYBLOB_EC;
+            if (!privkey.key) {
                 exit(ERR_CRYPTO);
             }
-            key_count++;
+        } else if (!strncmp(keyinfo, "rsa:", 4)) {
+            size = atoi(&keyinfo[4]);
+            if ((size < RSA_MIN) || (size > RSA_MAX)) {
+                log0(0, 0, 0, "Invalid RSA key size: %s", &keyinfo[4]);
+                exit(ERR_PARAM);
+            }
+            privkey.rsa = gen_RSA_key(size, RSA_EXP, keyfile);
+            privkey_type = KEYBLOB_RSA;
+            if (!privkey.key) {
+                exit(ERR_CRYPTO);
+            }
+        } else {
+            log0(0, 0, 0, "Invalid keyinfo: %s", keyinfo);
+            exit(ERR_PARAM);
+        }
+    } else {
+        privkey = read_private_key(keyfile, &privkey_type);
+        if (privkey_type == 0) {
+            exit(ERR_CRYPTO);
         }
     }
 #endif
@@ -367,7 +360,8 @@ void create_sockets(void)
         log0(0, 0, 0, "Error getting bind address: %s", gai_strerror(rval));
         exit(ERR_SOCKET);
     }
-    if (bind(listener, ai_rval->ai_addr, ai_rval->ai_addrlen) == SOCKET_ERROR) {
+    if (bind(listener, ai_rval->ai_addr,
+            (socklen_t)ai_rval->ai_addrlen) == SOCKET_ERROR) {
         sockerror(0, 0, 0, "Error binding socket for listener");
         closesocket(listener);
         exit(ERR_SOCKET);
@@ -500,16 +494,22 @@ void create_sockets(void)
         }
     }
     for (i = 0; i < pub_multi_count; i++) {
-        if (server_count > 0) {
+        if (use_ssm) {
             log3(0, 0, 0, "joining ssm for server IPs");
             if (!multicast_join(listener, 0, &pub_multi[i], m_interface,
-                                interface_count, server_keys, server_count)) {
+                                interface_count, server_list, server_count)) {
                 exit(ERR_SOCKET);
             }
-            if (has_proxy) {
+            if (proxy_count) {
+                if (!multicast_join(listener, 0, &pub_multi[i], m_interface,
+                                    interface_count, proxy_list, proxy_count)) {
+                    exit(ERR_SOCKET);
+                }
+            }
+            if (has_v4_proxy) {
                 log3(0, 0, 0, "joining ssm for proxy IPs");
                 if (!multicast_join(listener, 0, &pub_multi[i], m_interface,
-                                    interface_count, &proxy_info, 1)) {
+                                    interface_count, &proxy4_info, 1)) {
                     exit(ERR_SOCKET);
                 }
             }
@@ -531,10 +531,8 @@ int dirs_movable(const char *dir1, const char *dir2)
     char tempf1[MAXPATHNAME], tempf2[MAXPATHNAME];
     int fd;
 
-    snprintf(tempf1, sizeof(tempf1)-1, "%s%c_uftptmp1", dir1, PATH_SEP);
-    tempf1[sizeof(tempf1)-1] = '\x0';
-    snprintf(tempf2, sizeof(tempf1)-1, "%s%c_uftptmp2", dir2, PATH_SEP);
-    tempf2[sizeof(tempf2)-1] = '\x0';
+    snprintf(tempf1, sizeof(tempf1), "%s%c_uftptmp1", dir1, PATH_SEP);
+    snprintf(tempf2, sizeof(tempf1), "%s%c_uftptmp2", dir2, PATH_SEP);
     if ((fd = open(tempf1, O_WRONLY | O_CREAT, 0644)) < 0) {
         fprintf(stderr, "couldn't write to directory %s: %s\n",
                          dir1, strerror(errno));
@@ -619,8 +617,7 @@ void initialize(void)
                             "for dest directory\n");
             exit(ERR_PARAM);
         }
-        snprintf(tempf1, sizeof(tempf1)-1, "%s%c_uftptmp1",destdir[i],PATH_SEP);
-        tempf1[sizeof(tempf1)-1] = '\x0';
+        snprintf(tempf1, sizeof(tempf1), "%s%c_uftptmp1", destdir[i], PATH_SEP);
         if ((fd = open(tempf1, O_WRONLY | O_CREAT, 0644)) < 0) {
             perror("couldn't write to dest directory");
             exit(ERR_PARAM);
