@@ -69,7 +69,7 @@ struct timeval *getrecenttimeout(void)
     struct timeval current_timestamp, min_timestamp;
     int i, found_timeout, done, sent_naks;
     struct group_list_t *group;
-    unsigned int section, nak_count;
+    unsigned int section, nak_count, retry_naks;
     unsigned char *naks;
 
     gettimeofday(&current_timestamp, NULL);
@@ -82,6 +82,15 @@ struct timeval *getrecenttimeout(void)
             if (group->group_id != 0) {
                 if (cmptimestamp(current_timestamp, group->timeout_time) >= 0) {
                     switch (group->phase) {
+                    case PHASE_PREREGISTER:
+                        if (cmptimestamp(current_timestamp,
+                                         group->expire_time) >= 0) {
+                            glog1(group, "PROXY_KEY not received");
+                            send_abort(group, "PROXY_KEY not received");
+                        } else {
+                            set_timeout(group, 0);
+                        }
+                        break;
                     case PHASE_REGISTERED:
                         send_register(group);
                         break;
@@ -112,38 +121,40 @@ struct timeval *getrecenttimeout(void)
                     group->fileinfo.nak_time.tv_usec = 0;
                     // Send NAKs
                     sent_naks = 0;
-                retry_naks:
-                    for (section = group->fileinfo.nak_section_first;
-                            section < group->fileinfo.nak_section_last;
-                            section++) {
-                        naks = NULL;
-                        nak_count = get_naks(group, section, &naks);
-                        glog3(group, "read %d NAKs for section %d",
-                                     nak_count, section);
-                        if (nak_count > 0) {
-                            send_status(group, section, naks, nak_count);
-                            sent_naks = 1;
+                    do {
+                        retry_naks = 0;
+                        for (section = group->fileinfo.nak_section_first;
+                                section < group->fileinfo.nak_section_last;
+                                section++) {
+                            naks = NULL;
+                            nak_count = get_naks(group, section, &naks);
+                            glog3(group, "read %d NAKs for section %d",
+                                         nak_count, section);
+                            if (nak_count > 0) {
+                                send_status(group, section, naks, nak_count);
+                                sent_naks = 1;
+                            }
+                            free(naks);
+                            naks = NULL;
                         }
-                        free(naks);
-                        naks = NULL;
-                    }
-                    if (file_done(group, 1)) {
-                        glog2(group, "File transfer complete");
-                        send_complete(group, 0);
-                        if (group->phase != PHASE_MIDGROUP) {
-                            print_result_status(group);
+                        if (file_done(group, 1)) {
+                            glog2(group, "File transfer complete");
+                            send_complete(group, 0);
+                            if (group->phase != PHASE_MIDGROUP) {
+                                print_result_status(group);
+                            }
+                            file_cleanup(group, 0);
+                        } else if (group->fileinfo.got_done && !sent_naks) {
+                            // We didn't send any NAKs since the last time
+                            // but the server is asking for some,
+                            // so check all prior sections
+                            group->fileinfo.nak_section_last = 
+                                    group->fileinfo.nak_section_first;
+                            group->fileinfo.nak_section_first = 0;
+                            group->fileinfo.got_done = 0;
+                            retry_naks = 1;
                         }
-                        file_cleanup(group, 0);
-                    } else if (group->fileinfo.got_done && !sent_naks) {
-                        // We didn't send any NAKs since the last time
-                        // but the server is asking for some,
-                        // so check all prior sections
-                        group->fileinfo.nak_section_last = 
-                                group->fileinfo.nak_section_first;
-                        group->fileinfo.nak_section_first = 0;
-                        group->fileinfo.got_done = 0;
-                        goto retry_naks;
-                    }
+                    } while (retry_naks);
                 } else if ((group->fileinfo.nak_time.tv_sec != 0) &&
                            ((!found_timeout) ||
                             (cmptimestamp(group->fileinfo.nak_time,
@@ -173,7 +184,7 @@ struct timeval *getrecenttimeout(void)
             }
         }
         // Check timeout for proxy key request
-        if (has_proxy && (proxy_pubkey.key == 0)) {
+        if (has_v4_proxy && (proxy4_pubkey.key == 0)) {
             if (cmptimestamp(current_timestamp, next_keyreq_time) >= 0) {
                 send_key_req();
                 done = 0;
@@ -199,10 +210,15 @@ struct timeval *getrecenttimeout(void)
     }
     if (found_timeout) {
         tv = diff_timeval(min_timestamp, current_timestamp);
-        return &tv;
     } else {
-        return NULL;
+        tv.tv_sec = 1;
+        tv.tv_usec = 0;
     }
+    if (tv.tv_sec >= 1) {
+        tv.tv_sec = 1;
+        tv.tv_usec = 0;
+    }
+    return &tv;
 }
 
 /**
@@ -216,7 +232,7 @@ void mainloop(void)
     unsigned char *buf, *decrypted, *message;
     char rxname[INET6_ADDRSTRLEN];
     unsigned int decryptlen, meslen;
-    int packetlen, rval, i, ecn;
+    int packetlen, rval, ecn;
     uint8_t version, *func, tos;
     uint16_t txseq;
     union sockaddr_u src;
@@ -224,16 +240,14 @@ void mainloop(void)
     double new_grtt;
 
     log2(0, 0, 0, "%s", VERSIONSTR);
-    for (i = 0; i < key_count; i++) {
-        if (privkey_type[i] == KEYBLOB_RSA) {
-            log2(0, 0, 0, "Loaded %d bit RSA key with fingerprint %s",
-                  RSA_keylen(privkey[i].rsa) * 8,
-                  print_key_fingerprint(privkey[i], KEYBLOB_RSA));
-        } else {
-            log2(0, 0, 0, "Loaded ECDSA key with curve %s and fingerprint %s",
-                  curve_name(get_EC_curve(privkey[i].ec)),
-                  print_key_fingerprint(privkey[i], KEYBLOB_EC));
-        }
+    if (privkey_type == KEYBLOB_RSA) {
+        log2(0, 0, 0, "Loaded %d bit RSA key with fingerprint %s",
+              RSA_keylen(privkey.rsa) * 8,
+              print_key_fingerprint(privkey, KEYBLOB_RSA));
+    } else {
+        log2(0, 0, 0, "Loaded ECDSA key with curve %s and fingerprint %s",
+              curve_name(get_EC_curve(privkey.ec)),
+              print_key_fingerprint(privkey, KEYBLOB_EC));
     }
 
     buf = safe_calloc(MAXMTU, 1);
@@ -241,6 +255,11 @@ void mainloop(void)
     header = (struct uftp_h *)buf;
 
     while (1) {
+        if (user_abort) {
+            free(buf);
+            free(decrypted);
+            exit(ERR_INTERRUPTED);
+        }
         tv = getrecenttimeout();
         if (tv) {
             log5(0, 0, 0, "read timeout: %d.%06d", tv->tv_sec, tv->tv_usec);
@@ -256,7 +275,8 @@ void mainloop(void)
             log1(0, 0, 0, "getnameinfo failed: %s", gai_strerror(rval));
         }
 
-        if (header->version == UFTP_VER_NUM) {
+        if ((header->version == UFTP_VER_NUM) ||
+                (header->version == UFTP4_VER_NUM)) {
             version = header->version;
             group = find_group(ntohl(header->group_id), header->group_inst);
         } else {
@@ -294,10 +314,8 @@ void mainloop(void)
             }
 
             if (!validate_and_decrypt(buf, packetlen, &decrypted, &decryptlen,
-                    group->keytype, group->groupkey, group->groupsalt,
-                    group->ivlen, group->hashtype, group->grouphmackey,
-                    group->hmaclen, group->sigtype, group->keyextype,
-                    group->server_pubkey, group->server_pubkeylen)) {
+                                      group->keytype, group->s_app_key,
+                                      group->s_app_iv, group->ivlen)) {
                 glog1(group, "Rejecting message from %s: "
                              "decrypt/validate failed", rxname);
                 continue;
@@ -310,7 +328,8 @@ void mainloop(void)
                     ((header->func == FILEINFO) || (header->func == FILESEG) ||
                      (header->func == DONE) || (header->func == DONE_CONF) ||
                      (header->func == CONG_CTRL) || ((header->func == ABORT) &&
-                         (group->phase != PHASE_REGISTERED)))) {
+                         (group->phase != PHASE_REGISTERED) && 
+                         (group->phase != PHASE_PREREGISTER)))) {
                 glog1(group, "Rejecting %s message from %s: not encrypted",
                              func_name(header->func), rxname);
                 continue;
@@ -330,13 +349,13 @@ void mainloop(void)
             glog5(group, "grtt: %.3f", group->grtt);
         }
 
-        if (header->func == PROXY_KEY) {
-            handle_proxy_key(&src, message, meslen);
+        if ((version == UFTP4_VER_NUM) && (header->func == PROXY_KEY)) {
+            handle_v4_proxy_key(&src, message, meslen);
             continue;
         }
         if (header->func == HB_RESP) {
             handle_hb_response(listener, &src, message, meslen, hb_hosts,
-                               hbhost_count, privkey[0], privkey_type[0], uid);
+                               hbhost_count, privkey, privkey_type, uid);
             continue;
         }
         if (header->func == ANNOUNCE) {
@@ -357,7 +376,8 @@ void mainloop(void)
                 glog1(group, "Version mismatch");
                 continue;
             }
-            if (group->src_id != header->src_id) {
+            if ((group->src_id != header->src_id) &&
+                    (group->proxy_id != header->src_id)) {
                 glog1(group, "Source ID mismatch");
                 continue;
             }
@@ -366,10 +386,30 @@ void mainloop(void)
                 continue;
             }
             switch (group->phase) {
+            case PHASE_PREREGISTER:
+                if (*func == PROXY_KEY && 
+                        group->version == UFTP_VER_NUM) {
+                    handle_proxy_key(group, &src, header->src_id, 
+                                     message, meslen);
+                } else {
+                    glog1(group, "Expected PROXY_KEY, got %s",func_name(*func));
+                }
+                break;
             case PHASE_REGISTERED:
                 if (group->keytype != KEY_NONE) {
-                    if (*func == KEYINFO) {
-                        handle_keyinfo(group, message, meslen, header->src_id);
+                    if (*func == KEYINFO && group->version == UFTP_VER_NUM) {
+                        handle_keyinfo(group, buf, packetlen);
+                    } else if ((*func == KEYINFO) && 
+                               (group->version == UFTP4_VER_NUM)) {
+                        handle_v4_keyinfo(group, message, meslen, 
+                                          header->src_id);
+                    } else if (*func == PROXY_KEY && 
+                            group->version == UFTP_VER_NUM) {
+                        // silently ignore and reset timeout
+                        set_timeout(group, 0);
+                    } else if (*func == REG_CONF) {
+                        // could come from server if proxy is in use, ignore
+                        set_timeout(group, 0);
                     } else {
                         glog1(group, "Expected KEYINFO, got %s",
                                      func_name(*func));
@@ -388,8 +428,12 @@ void mainloop(void)
             case PHASE_MIDGROUP:
                 if (*func == FILEINFO) {
                     handle_fileinfo(group, message, meslen, rxtime);
-                } else if (*func == KEYINFO) {
-                    handle_keyinfo(group, message, meslen, header->src_id);
+                } else if ((*func == KEYINFO) &&
+                           (group->version == UFTP4_VER_NUM)) {
+                    handle_v4_keyinfo(group, message, meslen, header->src_id);
+                } else if ((*func == KEYINFO) &&
+                           (group->version == UFTP_VER_NUM)) {
+                    handle_keyinfo(group, buf, packetlen);
                 } else if (*func == DONE) {
                     handle_done(group, message, meslen);
                 } else {
