@@ -1,7 +1,7 @@
 /*
  *  UFTP - UDP based FTP with multicast
  *
- *  Copyright (C) 2001-2019   Dennis A. Bush, Jr.   bush@tcnj.edu
+ *  Copyright (C) 2001-2020   Dennis A. Bush, Jr.   bush@tcnj.edu
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -320,7 +320,7 @@ void set_tfmcc_ack_info(struct group_list_t *group,
                         struct tfmcc_ack_info_he *tfmcc)
 {
     struct timeval now, send_time;
-    int64_t ccrate;
+    int64_t ccrate, send_time_us;
 
     tfmcc->exttype = EXT_TFMCC_ACK_INFO;
     tfmcc->extlen = sizeof(struct tfmcc_ack_info_he) / 4;
@@ -345,8 +345,14 @@ void set_tfmcc_ack_info(struct group_list_t *group,
         send_time = add_timeval(group->last_server_ts,
                 diff_timeval(now, group->last_server_rx_ts));
     }
-    tfmcc->tstamp_sec = htonl((uint32_t)send_time.tv_sec);
-    tfmcc->tstamp_usec = htonl((uint32_t)send_time.tv_usec);
+    if (group->version == UFTP4_VER_NUM) {
+        tfmcc->tstamp_hi = htonl((uint32_t)send_time.tv_sec);
+        tfmcc->tstamp_lo = htonl((uint32_t)send_time.tv_usec);
+    } else {
+        send_time_us = tv_to_usec(send_time);
+        tfmcc->tstamp_hi = htonl((send_time_us & 0xFFFFFFFF00000000ULL) >> 32);
+        tfmcc->tstamp_lo = htonl(send_time_us & 0x00000000FFFFFFFFULL);
+    }
 }
 
 /**
@@ -394,10 +400,8 @@ void send_status(struct group_list_t *group, unsigned int section,
     if ((group->phase != PHASE_REGISTERED) && (group->keytype != KEY_NONE)) {
         encrypted = NULL;
         if (!encrypt_and_sign(buf, &encrypted, payloadlen, &enclen,
-                group->keytype, group->groupkey, group->groupsalt,&group->ivctr,
-                group->ivlen, group->hashtype, group->grouphmackey,
-                group->hmaclen, group->sigtype, group->keyextype,
-                group->client_privkey, group->client_privkeylen)) {
+                group->keytype, group->c_app_key, group->c_app_iv,
+                &group->ivctr, group->ivlen)) {
             glog0(group, "Error encrypting STATUS");
             free(buf);
             return;
@@ -449,11 +453,11 @@ void print_result_status(struct group_list_t *group)
     t = time(NULL);
     done_time = localtime(&t);
     fprintf(status_file,
-            "RESULT;%04d/%02d/%02d-%02d:%02d:%02d;%08X;%08X;%s;%sKB",
+            "RESULT;%04d/%02d/%02d-%02d:%02d:%02d;%08X;%08X;%s;" F_u64 "KB",
             done_time->tm_year + 1900, done_time->tm_mon + 1,
             done_time->tm_mday, done_time->tm_hour, done_time->tm_min,
             done_time->tm_sec, ntohl(group->src_id), group->group_id,
-            group->fileinfo.name, printll(group->fileinfo.size / 1024));
+            group->fileinfo.name, group->fileinfo.size / 1024);
     switch (group->fileinfo.comp_status) {
     case COMP_STAT_NORMAL:
         fprintf(status_file, ";copied\n");
@@ -519,10 +523,8 @@ void send_complete(struct group_list_t *group, int set_freespace)
     if ((group->phase != PHASE_REGISTERED) && (group->keytype != KEY_NONE)) {
         encrypted = NULL;
         if (!encrypt_and_sign(buf, &encrypted, payloadlen, &enclen,
-                group->keytype, group->groupkey, group->groupsalt,&group->ivctr,
-                group->ivlen, group->hashtype, group->grouphmackey,
-                group->hmaclen, group->sigtype, group->keyextype,
-                group->client_privkey, group->client_privkeylen)) {
+                group->keytype, group->c_app_key, group->c_app_iv,
+                &group->ivctr, group->ivlen)) {
             glog0(group, "Error encrypting COMPLETE");
             free(buf);
             return;
@@ -576,10 +578,8 @@ void send_cc_ack(struct group_list_t *group)
     if ((group->phase != PHASE_REGISTERED) && (group->keytype != KEY_NONE)) {
         encrypted = NULL;
         if (!encrypt_and_sign(buf, &encrypted, payloadlen, &enclen,
-                group->keytype, group->groupkey, group->groupsalt,&group->ivctr,
-                group->ivlen, group->hashtype, group->grouphmackey,
-                group->hmaclen, group->sigtype, group->keyextype,
-                group->client_privkey, group->client_privkeylen)) {
+                group->keytype, group->c_app_key, group->c_app_iv,
+                &group->ivctr, group->ivlen)) {
             glog0(group, "Error encrypting CC_ACK");
             free(buf);
             return;
@@ -1033,7 +1033,7 @@ void handle_cong_ctrl(struct group_list_t *group, const unsigned char *message,
     const struct cong_ctrl_h *cong_ctrl;
     const struct cc_item *cc_list;
     int listlen, i, ccidx, clridx;
-    int64_t ccrate;
+    int64_t ccrate, time_us;
     uint16_t new_ccseq;
     double new_rtt;
 
@@ -1054,8 +1054,13 @@ void handle_cong_ctrl(struct group_list_t *group, const unsigned char *message,
 
     new_ccseq = ntohs(cong_ctrl->cc_seq);
     ccrate = unquantize_rate(ntohs(cong_ctrl->cc_rate));
-    group->last_server_ts.tv_sec = ntohl(cong_ctrl->tstamp_sec);
-    group->last_server_ts.tv_usec = ntohl(cong_ctrl->tstamp_usec);
+    if (group->version == UFTP4_VER_NUM) {
+        group->last_server_ts.tv_sec = ntohl(cong_ctrl->tstamp_hi);
+        group->last_server_ts.tv_usec = ntohl(cong_ctrl->tstamp_lo);
+    } else {
+        time_us = (int64_t)ntohl(cong_ctrl->tstamp_hi) << 32;
+        time_us |= ntohl(cong_ctrl->tstamp_lo);
+    }
     group->last_server_rx_ts = rxtime;
 
     for (clridx = -1, ccidx = -1, i = 0;
