@@ -1,7 +1,7 @@
 /*
  *  UFTP - UDP based FTP with multicast
  *
- *  Copyright (C) 2001-2018   Dennis A. Bush, Jr.   bush@tcnj.edu
+ *  Copyright (C) 2001-2020   Dennis A. Bush, Jr.   bush@tcnj.edu
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -130,31 +130,30 @@ void group_cleanup(struct pr_group_list_t *group)
                 m_interface, interface_count, server_fp, server_fp_count);
     }
     if (group->server_pubkey.key) {
-        if ((group->keyextype == KEYEX_RSA) ||
-                (group->keyextype == KEYEX_ECDH_RSA)) {
+        if (group->server_pubkeytype == KEYBLOB_RSA) {
             free_RSA_key(group->server_pubkey.rsa);
         } else {
             free_EC_key(group->server_pubkey.ec);
         }
-        if ((group->keyextype == KEYEX_ECDH_RSA) ||
-                (group->keyextype == KEYEX_ECDH_ECDSA)) {
-            free_EC_key(group->server_dhkey.ec);
-            free_EC_key(group->proxy_dhkey.ec);
-        }
     }
+    free_EC_key(group->server_dhkey.ec);
+    free_EC_key(group->proxy_u_dhkey.ec);
+    free_EC_key(group->proxy_d_dhkey.ec);
+    free(group->s_context);
+    free(group->p_context);
+    free(group->c_context1);
+    free(group->c_context2);
     for (i = 0; i < group->destcount; i++) {
-        if (group->destinfo[i].pubkey.key) {
-            if ((group->keyextype == KEYEX_RSA) ||
-                    (group->keyextype == KEYEX_ECDH_RSA)) {
-                free_RSA_key(group->destinfo[i].pubkey.rsa);
+        if (group->destinfo[i].client_pubkey.key) {
+            if (group->destinfo[i].client_pubkeytype == KEYBLOB_RSA) {
+                free_RSA_key(group->destinfo[i].client_pubkey.rsa);
             } else {
-                free_EC_key(group->destinfo[i].pubkey.ec);
-            }
-            if ((group->keyextype == KEYEX_ECDH_RSA) ||
-                    (group->keyextype == KEYEX_ECDH_ECDSA)) {
-                free_EC_key(group->destinfo[i].dhkey.ec);
+                free_EC_key(group->destinfo[i].client_pubkey.ec);
             }
         }
+        free_EC_key(group->destinfo[i].client_dhkey.ec);
+        free(group->destinfo[i].c_context1);
+        free(group->destinfo[i].c_context2);
     }
     memset(group, 0, sizeof(struct pr_group_list_t));
 }
@@ -178,9 +177,9 @@ void set_uftp_header(struct uftp_h *header, int func,
     case COMPLETE:
         header->seq = htons(group->send_seq_up++);
         break;
+    case PROXY_KEY:
     case KEYINFO:
         header->seq = 0;
-        header->src_id = group->src_id;
         header->grtt = quantize_grtt(group->grtt);
         header->gsize = quantize_gsize(group->gsize);
         break;
@@ -326,6 +325,7 @@ void check_pending(struct pr_group_list_t *group, int hostidx,
     const uint8_t *func;
     struct pr_pending_info_t *pending;
     int match, pendidx, hlen;
+    int64_t tstamp_usec;
 
     func = message;
     fileinfoack = (const struct fileinfoack_h *)message;
@@ -397,8 +397,14 @@ void check_pending(struct pr_group_list_t *group, int hostidx,
         if (pending->count == 1) {
             pending->partial = 1;
             gettimeofday(&pending->rx_tstamp, NULL);
-            pending->tstamp.tv_sec = ntohl(fileinfoack->tstamp_sec);
-            pending->tstamp.tv_usec = ntohl(fileinfoack->tstamp_usec);
+            if (group->version == UFTP4_VER_NUM) {
+                pending->tstamp.tv_sec = ntohl(fileinfoack->tstamp_hi);
+                pending->tstamp.tv_usec = ntohl(fileinfoack->tstamp_lo);
+            } else {
+                tstamp_usec = (int64_t)ntohl(fileinfoack->tstamp_hi) << 32;
+                tstamp_usec |= ntohl(fileinfoack->tstamp_lo);
+                pending->tstamp = usec_to_tv(tstamp_usec);
+            }
             glog3(group, "send time = %d.%06d",
                          pending->tstamp.tv_sec, pending->tstamp.tv_usec);
             glog3(group, "rx time = %d.%06d",
@@ -512,7 +518,6 @@ void forward_message(struct pr_group_list_t *group,
                      unsigned char *packet, int packetlen)
 {
     struct uftp_h *header;
-    struct encrypted_h *encrypted;
     struct announce_h *announce;
     struct enc_info_he *encinfo;
     union sockaddr_u dest;
@@ -520,7 +525,6 @@ void forward_message(struct pr_group_list_t *group,
     int hostidx, rval, iplen, resign;
     char destname[INET6_ADDRSTRLEN], destport[PORTNAME_LEN];
     uint8_t *sig, *sigcopy;
-    union key_t key;
 
     header = (struct uftp_h *)packet;
     meslen = (unsigned int)packetlen;
@@ -539,7 +543,6 @@ void forward_message(struct pr_group_list_t *group,
             } else {
                 dest = group->privatemcast;
             }
-            key = group->server_pubkey;
         }
     } else {
         dest = group->up_addr;
@@ -549,22 +552,14 @@ void forward_message(struct pr_group_list_t *group,
                 glog1(group, "Couldn't find receiver in list");
                 return;
             }
-            key = group->destinfo[hostidx].pubkey;
         }
     }
 
-    // If we're using KEYEX signatures, or sending an ANNOUNCE with ECDH,
+    // If we're sending an ANNOUNCE with ECDH under V4,
     // verify the signature and resign
     resign = 0;
-    if ((proxy_type != SERVER_PROXY) && (header->func == ENCRYPTED) &&
-            (group->sigtype == SIG_KEYEX)) {
-        encrypted = (struct encrypted_h *)(packet + sizeof(struct uftp_h));
-        sig = (uint8_t *)encrypted + sizeof(struct encrypted_h);
-        siglen = ntohs(encrypted->sig_len);
-        resign = 1;
-    } else if ((proxy_type != SERVER_PROXY) && (header->func == ANNOUNCE) &&
-            ((group->keyextype == KEYEX_ECDH_RSA) ||
-             (group->keyextype == KEYEX_ECDH_ECDSA))) {
+    if ((proxy_type != SERVER_PROXY) && (header->func == ANNOUNCE) &&
+            (group->keytype != KEY_NONE) && (group->version == UFTP4_VER_NUM)) {
         announce = (struct announce_h *)(packet + sizeof(struct uftp_h));
         iplen = ((announce->flags & FLAG_IPV6) != 0) ? 16 : 4;
         encinfo = (struct enc_info_he *) ((uint8_t *)announce +
@@ -578,16 +573,7 @@ void forward_message(struct pr_group_list_t *group,
         sigcopy = safe_calloc(siglen, 1);
         memcpy(sigcopy, sig, siglen);
         memset(sig, 0, siglen);
-        if ((group->keyextype == KEYEX_RSA) ||
-                (group->keyextype == KEYEX_ECDH_RSA)) {
-            if (header->func == ENCRYPTED) {
-                if (!verify_RSA_sig(key.rsa, group->hashtype, packet,
-                                    meslen, sigcopy, siglen)) {
-                    glog1(group, "Signature verification failed");
-                    free(sigcopy);
-                    return;
-                }
-            }
+        if (group->proxy_privkeytype == KEYBLOB_RSA) {
             if (!create_RSA_sig(group->proxy_privkey.rsa, group->hashtype,
                                 packet, meslen, sigcopy, &siglen)) {
                 glog0(group, "Signature creation failed");
@@ -595,14 +581,6 @@ void forward_message(struct pr_group_list_t *group,
                 return;
             }
         } else {
-            if (header->func == ENCRYPTED) {
-                if (!verify_ECDSA_sig(key.ec, group->hashtype, packet,
-                                      meslen, sigcopy, siglen)) {
-                    glog1(group, "Signature verification failed");
-                    free(sigcopy);
-                    return;
-                }
-            }
             if (!create_ECDSA_sig(group->proxy_privkey.ec, group->hashtype,
                                   packet, meslen, sigcopy, &siglen)) {
                 glog0(group, "Signature creation failed");
@@ -677,6 +655,7 @@ void handle_hb_request(const union sockaddr_u *src,
                 if (memcmp(down_fingerprint, fingerprint, fplen)) {
                     log1(0, 0, 0, "Failed to verify HB_REQ fingerprint");
                     resp = HB_AUTH_FAILED;
+                    free_RSA_key(key.rsa);
                     goto end;
                 }
 
@@ -685,8 +664,10 @@ void handle_hb_request(const union sockaddr_u *src,
                         sizeof(hbreq->nonce), sig, siglen)) {
                     log1(0, 0, 0, "Failed to verify HB_REQ signature");
                     resp = HB_AUTH_FAILED;
+                    free_RSA_key(key.rsa);
                     goto end;
                 }
+                free_RSA_key(key.rsa);
             } else {
                 if (!import_EC_key(&key.ec, keyblob, bloblen, 0)) {
                     log1(0, 0, 0, "Failed to import public key from HB_REQ");
@@ -698,6 +679,7 @@ void handle_hb_request(const union sockaddr_u *src,
                 if (memcmp(down_fingerprint, fingerprint, fplen)) {
                     log1(0, 0, 0, "Failed to verify HB_REQ fingerprint");
                     resp = HB_AUTH_FAILED;
+                    free_EC_key(key.ec);
                     goto end;
                 }
 
@@ -706,8 +688,10 @@ void handle_hb_request(const union sockaddr_u *src,
                         sizeof(hbreq->nonce), sig, siglen)) {
                     log1(0, 0, 0, "Failed to verify HB_REQ signature");
                     resp = HB_AUTH_FAILED;
+                    free_EC_key(key.ec);
                     goto end;
                 }
+                free_EC_key(key.ec);
             }
 
             down_addr = *src;
@@ -730,12 +714,18 @@ end:
 void handle_key_req(const union sockaddr_u *src,
                     const unsigned char *packet, unsigned packetlen)
 {
+    const struct uftp_h *header;
     const struct key_req_h *keyreq;
     struct timeval current_timestamp;
     char destname[INET6_ADDRSTRLEN];
     int rval;
 
+    header = (const struct uftp_h *)packet;
     keyreq = (const struct key_req_h *)(packet + sizeof(struct uftp_h));
+    if (header->version != UFTP4_VER_NUM) {
+        log1(0, 0, 0, "invalid version for KEY_REQ");
+        return;
+    }
 
     if ((rval = getnameinfo((const struct sockaddr *)src, family_len(*src),
             destname, sizeof(destname), NULL, 0, NI_NUMERICHOST)) != 0) {
@@ -750,7 +740,7 @@ void handle_key_req(const union sockaddr_u *src,
 
     gettimeofday(&current_timestamp, NULL);
     if (diff_sec(current_timestamp, last_key_req) > KEY_REQ_LIMIT) {
-        send_proxy_key();
+        send_v4_proxy_key();
     }
 }
 
@@ -797,24 +787,23 @@ void send_hb_response(const union sockaddr_u *src, int response)
 /**
  * Sends a PROXY_KEY message to each public multicast address.
  */
-void send_proxy_key()
+void send_v4_proxy_key()
 {
     unsigned char *packet, *keyblob, *dhblob, *sig;
     struct uftp_h *header;
-    struct proxy_key_h *proxykey;
+    struct proxy_key_4_h *proxykey;
     uint32_t nonce;
     unsigned int meslen, siglen;
     uint16_t bloblen, dhlen;
     char pubname[INET6_ADDRSTRLEN];
     int rval, i;
 
-    packet = safe_calloc(sizeof(struct uftp_h) + sizeof(struct hb_req_h) +
-                    (PUBKEY_LEN * 3) , 1);
+    packet = safe_calloc(MAXMTU, 1);
 
     header = (struct uftp_h *)packet;
-    proxykey = (struct proxy_key_h *)(packet + sizeof(struct uftp_h));
-    keyblob = (unsigned char *)proxykey + sizeof(struct proxy_key_h);
-    header->version = UFTP_VER_NUM;
+    proxykey = (struct proxy_key_4_h *)(packet + sizeof(struct uftp_h));
+    keyblob = (unsigned char *)proxykey + sizeof(struct proxy_key_4_h);
+    header->version = UFTP4_VER_NUM;
     header->func = PROXY_KEY;
     header->src_id = uid;
     proxykey->func = PROXY_KEY;
@@ -835,8 +824,8 @@ void send_proxy_key()
         }
     }
     dhblob = keyblob + bloblen;
-    if (dhkey.key) {
-        if (!export_EC_key(dhkey.ec, dhblob, &dhlen)) {
+    if (v4_dhkey.key) {
+        if (!export_EC_key(v4_dhkey.ec, dhblob, &dhlen)) {
             log0(0, 0, 0, "Error exporting public key");
             free(packet);
             return;
@@ -863,21 +852,22 @@ void send_proxy_key()
     proxykey->bloblen = htons(bloblen);
     proxykey->dhlen = htons(dhlen);
     proxykey->siglen = htons(siglen);
-    proxykey->hlen = (sizeof(struct proxy_key_h) + bloblen + dhlen + siglen)/4;
+    proxykey->hlen = (uint8_t)((sizeof(struct proxy_key_4_h) +
+                        bloblen + dhlen + siglen) / 4);
 
     meslen = sizeof(struct uftp_h) + (proxykey->hlen * 4);
     for (i = 0; i < pub_multi_count; i++) {
         if (nb_sendto(listener, packet, meslen, 0, 
                       (struct sockaddr *)&pub_multi[i],
                       family_len(pub_multi[i])) == SOCKET_ERROR) {
-            sockerror(0, 0, 0, "Error sending PROXY_KEY");
+            sockerror(0, 0, 0, "Error sending PROXY_KEY_4");
         } else {
             if ((rval = getnameinfo((struct sockaddr *)&pub_multi[i],
                     family_len(pub_multi[i]), pubname, sizeof(pubname), NULL, 0,
                     NI_NUMERICHOST)) != 0) {
                 log1(0, 0, 0, "getnameinfo failed: %s", gai_strerror(rval));
             }
-            log2(0, 0, 0, "Sent PROXY_KEY to %s", pubname);
+            log2(0, 0, 0, "Sent PROXY_KEY_4 to %s", pubname);
         }
     }
     free(packet);
@@ -942,6 +932,8 @@ void send_downstream_abort(struct pr_group_list_t *group, uint32_t dest_id,
     set_uftp_header(header, ABORT, group);
     header->seq = group->send_seq_down++;
     header->src_id = uid;
+    header->grtt = quantize_grtt(group->grtt);
+    header->gsize = quantize_gsize(group->gsize);
     abort_hdr->func = ABORT;
     abort_hdr->hlen = sizeof(struct abort_h) / 4;
     if ((dest_id == 0) && current) {
@@ -1051,14 +1043,14 @@ int verify_fingerprint(const struct fp_list_t *fplist, int listlen,
  * Returns the verify_data string used in certain messages.  This value
  * is then run through the PRF with the result going into the message.
  */
-uint8_t *build_verify_data(struct pr_group_list_t *group, int hostidx,
-                           int *verifylen, int full)
+uint8_t *build_v4_verify_data(struct pr_group_list_t *group, int hostidx,
+                              int *verifylen, int full)
 {
     uint8_t *verifydata, *keyblob;
     uint32_t group_id;
     struct pr_destinfo_t *dest;
     union key_t key;
-    int iplen;
+    int iplen, keytype;
     uint16_t bloblen;
 
     iplen = (group->privatemcast.ss.ss_family == AF_INET6) ?
@@ -1107,12 +1099,13 @@ uint8_t *build_verify_data(struct pr_group_list_t *group, int hostidx,
         if (group->client_auth) {
             if (hostidx == -1) {
                 key = group->proxy_privkey;
+                keytype = group->proxy_privkeytype;
             } else {
-                key = dest->pubkey;
+                key = dest->client_pubkey;
+                keytype = dest->client_pubkeytype;
             }
             keyblob = verifydata + *verifylen;
-            if ((group->keyextype == KEYEX_RSA) ||
-                    (group->keyextype == KEYEX_ECDH_RSA)) {
+            if (keytype == KEYBLOB_RSA) {
                 if (!export_RSA_key(key.rsa, keyblob, &bloblen)) {
                     free(verifydata);
                     return NULL;
