@@ -1,7 +1,7 @@
 /*
  *  UFTP - UDP based FTP with multicast
  *
- *  Copyright (C) 2001-2019   Dennis A. Bush, Jr.   bush@tcnj.edu
+ *  Copyright (C) 2001-2020   Dennis A. Bush, Jr.   bush@tcnj.edu
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -77,8 +77,13 @@ void cleanup(void)
     }
     if (!parent) {
         for (i = 0; i < pub_multi_count; i++) {
-            multicast_leave(listener, 0, &pub_multi[i], m_interface,
-                            interface_count, server_fp, server_fp_count);
+            if (use_ssm) {
+                multicast_leave(listener, 0, &pub_multi[i], m_interface,
+                                interface_count, server_fp, server_fp_count);
+            } else {
+                multicast_leave(listener, 0, &pub_multi[i], m_interface,
+                                interface_count, NULL, 0);
+            }
         }
     }
     closesocket(listener);
@@ -90,6 +95,7 @@ void cleanup(void)
             free_EC_key(privkey[i].ec);
         }
     }
+    free_EC_key(v4_dhkey.ec);
     crypto_cleanup();
 
 #ifdef WINDOWS
@@ -104,7 +110,7 @@ void cleanup(void)
 void gotsig(int sig)
 {
     log0(0, 0, 0, "Exiting on signal %d", sig);
-    exit(ERR_INTERRUPTED);
+    user_abort = 1;
 }
 
 #ifdef WINDOWS
@@ -133,7 +139,8 @@ BOOL WINAPI winsig(DWORD event)
         log0(0, 0, 0, "GOT unknown event %d", event);
         break;
     }
-    exit(ERR_INTERRUPTED);
+    user_abort = 1;
+    return TRUE;
 }
 #endif
 
@@ -257,11 +264,11 @@ void key_init(void)
     crypto_init(sys_keys);
 
     if ((keyfile_count == 0) && (keyinfo_count == 0)) {
-        privkey[0].rsa = gen_RSA_key(0, RSA_EXP, NULL);
+        privkey[0].ec = gen_EC_key(DEF_CURVE, 0, NULL);
         if (!privkey[0].key) {
             exit(ERR_CRYPTO);
         }
-        privkey_type[0] = KEYBLOB_RSA;
+        privkey_type[0] = KEYBLOB_EC;
         key_count = 1;
     } else if (keyinfo_count != 0) {
         key_count = 0;
@@ -284,7 +291,7 @@ void key_init(void)
                 }
             } else if (!strncmp(keyinfo[i], "rsa:", 4)) {
                 size = atoi(&keyinfo[i][4]);
-                if ((size < 512) || (size > 2048)) {
+                if ((size < RSA_MIN) || (size > RSA_MAX)) {
                     log0(0, 0, 0, "Invalid RSA key size: %s", &keyinfo[i][4]);
                     exit(ERR_PARAM);
                 }
@@ -309,9 +316,9 @@ void key_init(void)
             key_count++;
         }
     }
-    if ((proxy_type == RESPONSE_PROXY) && (ecdh_curve != 0)) {
-        dhkey.ec = gen_EC_key(ecdh_curve, 1, NULL);
-        if (!dhkey.key) {
+    if ((proxy_type == RESPONSE_PROXY) && (v4_ecdh_curve != 0)) {
+        v4_dhkey.ec = gen_EC_key(v4_ecdh_curve, 1, NULL);
+        if (!v4_dhkey.key) {
             log0(0, 0, 0, "Failed to generate DH key");
             exit(ERR_CRYPTO);
         }
@@ -405,7 +412,8 @@ void create_sockets(void)
         log0(0, 0, 0, "Error getting bind address: %s", gai_strerror(rval));
         exit(ERR_SOCKET);
     }
-    if (bind(listener, ai_rval->ai_addr, ai_rval->ai_addrlen) == SOCKET_ERROR) {
+    if (bind(listener, ai_rval->ai_addr, 
+            (socklen_t)ai_rval->ai_addrlen) == SOCKET_ERROR) {
         sockerror(0, 0, 0, "Error binding socket for listener");
         closesocket(listener);
         exit(ERR_SOCKET);
@@ -581,9 +589,16 @@ void create_sockets(void)
     }
 
     for (i = 0; i < pub_multi_count; i++) {
-        if (!multicast_join(listener, 0, &pub_multi[i], m_interface,
-                            interface_count, server_fp, server_fp_count)) {
-            exit(ERR_SOCKET);
+        if (use_ssm) {
+            if (!multicast_join(listener, 0, &pub_multi[i], m_interface,
+                                interface_count, server_fp, server_fp_count)) {
+                exit(ERR_SOCKET);
+            }
+        } else {
+            if (!multicast_join(listener, 0, &pub_multi[i], m_interface,
+                                interface_count, NULL, 0)) {
+                exit(ERR_SOCKET);
+            }
         }
     }
 }
