@@ -57,22 +57,21 @@
  */
 int set_enc_info(const struct finfo_t *finfo, struct enc_info_he *encinfo)
 {
-    unsigned char *keyblob;
-    uint16_t bloblen;
+    uint8_t *keyblob, *dhblob;
+    uint16_t bloblen, dhlen;
     int extlen;
 
     keyblob = ((uint8_t *)encinfo + sizeof(struct enc_info_he));
 
     encinfo->exttype = EXT_ENC_INFO;
-    encinfo->keyextype_sigtype = (keyextype & 0x0F) << 4;
-    encinfo->keyextype_sigtype |= (sigtype & 0x0F);
+    encinfo->keyextype_sigtype = 0;
     encinfo->keytype = keytype;
     encinfo->hashtype = hashtype;
     if (client_auth) {
         encinfo->flags |= FLAG_CLIENT_AUTH;
     }
     memcpy(encinfo->rand1, rand1, sizeof(rand1));
-    if ((keyextype == KEYEX_RSA) || (keyextype == KEYEX_ECDH_RSA)) {
+    if (privkeytype == KEYBLOB_RSA) {
         if (!export_RSA_key(privkey.rsa, keyblob, &bloblen)) {
             glog0(finfo, "Error exporting server public key");
             return 0;
@@ -84,29 +83,33 @@ int set_enc_info(const struct finfo_t *finfo, struct enc_info_he *encinfo)
         }
     }
     encinfo->keylen = htons(bloblen);
-    if ((keyextype == KEYEX_ECDH_RSA) || (keyextype == KEYEX_ECDH_ECDSA)) {
-        uint16_t dhlen;
-        uint8_t *dhblob = ((uint8_t *)encinfo + sizeof(struct enc_info_he) +
-                          ntohs(encinfo->keylen));
+    dhblob = ((uint8_t *)encinfo + sizeof(struct enc_info_he) +
+                                   ntohs(encinfo->keylen));
 
-        if (!export_EC_key(dhkey.ec, dhblob, &dhlen)) {
-            glog0(finfo, "Error exporting server ECDH public key");
-            return 0;
-        }
-        encinfo->dhlen = htons(dhlen);
-        if (keyextype == KEYEX_ECDH_RSA) {
-            encinfo->siglen = htons(RSA_keylen(privkey.rsa)); 
-        } else {
-            encinfo->siglen = htons(ECDSA_siglen(privkey.ec)); 
-        }
+    if (!export_EC_key(dhkey.ec, dhblob, &dhlen)) {
+        glog0(finfo, "Error exporting server ECDH public key");
+        return 0;
+    }
+    encinfo->dhlen = htons(dhlen);
+    if (privkeytype == KEYBLOB_RSA) {
+        encinfo->siglen = htons(RSA_keylen(privkey.rsa)); 
     } else {
-        encinfo->dhlen = 0;
-        encinfo->siglen = 0;
+        encinfo->siglen = htons(ECDSA_siglen(privkey.ec)); 
     }
     
     extlen = sizeof(struct enc_info_he) + ntohs(encinfo->keylen) +
-            ntohs(encinfo->dhlen) + ntohs(encinfo->siglen);
+             ntohs(encinfo->dhlen) + ntohs(encinfo->siglen);
     encinfo->extlen = extlen / 4;
+
+    if (!servercontext) {
+        create_server_context(htonl(finfo->group_id), finfo->group_inst,
+                              server_id, encinfo, extlen,
+                              &servercontext, &servercontext_len);
+        calculate_server_app_keys(hashtype, groupmaster, sizeof(groupmaster),
+                servercontext, servercontext_len,
+                keylen, ivlen, server_app_key, server_app_iv);
+    }
+
     return extlen;
 }
 
@@ -127,6 +130,7 @@ int send_announce(const struct finfo_t *finfo, int attempt, int open)
     struct enc_info_he *encinfo;
     struct timeval tv;
     uint32_t *idlist;
+    int64_t time_us;
 
     buf = safe_calloc(MAXMTU, 1); 
     if (listen_dest.ss.ss_family == AF_INET6) {
@@ -153,8 +157,9 @@ int send_announce(const struct finfo_t *finfo, int attempt, int open)
     announce->cc_type = cc_type;
     announce->blocksize = htons(blocksize);
     gettimeofday(&tv, NULL);
-    announce->tstamp_sec = htonl(tv.tv_sec);
-    announce->tstamp_usec = htonl(tv.tv_usec);
+    time_us = tv_to_usec(tv);
+    announce->tstamp_hi = htonl((time_us & 0xFFFFFFFF00000000ULL) >> 32);
+    announce->tstamp_lo = htonl(time_us & 0x00000000FFFFFFFFULL);
     if (!is_multicast(&listen_dest, 0)) {
         memset(publicaddr, 0, iplen);
         memset(privateaddr, 0, iplen);
@@ -176,10 +181,11 @@ int send_announce(const struct finfo_t *finfo, int attempt, int open)
             free(buf);
             return 0;
         }
-        announce->hlen = (sizeof(struct announce_h) +
-                          iplen + iplen + extlen) / 4;
+        announce->hlen = (uint8_t)((sizeof(struct announce_h) +
+                          iplen + iplen + extlen) / 4);
     } else {
-        announce->hlen = (sizeof(struct announce_h) + iplen + iplen) / 4;
+        announce->hlen = (uint8_t)((sizeof(struct announce_h) +
+                         iplen + iplen) / 4);
     }
 
     idlist = (uint32_t *)((uint8_t *)announce + (announce->hlen * 4));
@@ -249,65 +255,115 @@ int send_keyinfo(const struct finfo_t *finfo, int attempt)
     unsigned char *buf, *iv;
     struct uftp_h *header;
     struct keyinfo_h *keyinfo;
+    unsigned char *sig, *sigcopy, *sigcontext, *aadcontext;
     struct destkey *keylist;
-    unsigned int hsize, payloadlen, len;
-    int maxdest, packetcnt, dests, iv_init, i;
-    int unauth_keytype, unauth_keylen, unauth_ivlen;
-
-    // Don't use a cipher in an authentication mode to encrypt the group master
-    unauth_keytype = unauth_key(keytype);
-    get_key_info(unauth_keytype, &unauth_keylen, &unauth_ivlen);
+    unsigned int hsize, packetlen, enclen, siglen, _siglen;
+    unsigned int basesigcontextlen, sigcontextlen, baseaadlen, aadlen;
+    int maxdest, packetcnt, dests, iv_init, i, rval;
+    const char *sig_context_str = "UFTP 5, KEYINFO";
+    const char *aad_context_str = "UFTP 5, group master";
 
     buf = safe_calloc(MAXMTU, 1);
-    iv = safe_calloc(unauth_ivlen, 1);
+    iv = safe_calloc(ivlen, 1);
     header = (struct uftp_h *)buf;
     keyinfo = (struct keyinfo_h *)(buf + sizeof(struct uftp_h));
-    keylist = (struct destkey *)((char *)keyinfo + sizeof(struct keyinfo_h));
+
+    if (privkeytype == KEYBLOB_RSA) {
+        siglen = RSA_keylen(privkey.rsa);
+    } else {
+        siglen = ECDSA_siglen(privkey.ec);
+    }
 
     set_uftp_header(header, KEYINFO, finfo->group_id, finfo->group_inst,
                     grtt, destcount);
     keyinfo->func = KEYINFO;
-    keyinfo->hlen = sizeof(struct keyinfo_h) / 4;
-    keylist = (struct destkey *)((uint8_t *)keyinfo + (keyinfo->hlen * 4));
+    keyinfo->hlen = (uint8_t)((sizeof(struct keyinfo_h) + siglen) / 4);
+    keyinfo->siglen = htons(siglen);
+    sig = (uint8_t *)keyinfo + sizeof(struct keyinfo_h);
+    keylist = (struct destkey *)(sig + siglen);
+
+    sigcontext = safe_malloc(strlen(sig_context_str) + servercontext_len +
+                             MAXMTU);
+    aadcontext = safe_malloc(strlen(aad_context_str) + MAXMTU);
+    sigcopy = safe_malloc(siglen);
+    sigcontextlen = 0;
+    memcpy(sigcontext + sigcontextlen, sig_context_str, strlen(sig_context_str)); 
+    sigcontextlen += (unsigned)strlen(sig_context_str);
+    memcpy(sigcontext + sigcontextlen, servercontext, servercontext_len);
+    sigcontextlen += servercontext_len;
+    basesigcontextlen = sigcontextlen;
+    aadlen = 0;
+    memcpy(aadcontext + aadlen, aad_context_str, strlen(aad_context_str)); 
+    aadlen += (unsigned)strlen(aad_context_str);
+    baseaadlen = aadlen;
 
     iv_init = 0;
-    hsize = sizeof(struct keyinfo_h);
+    hsize = sizeof(struct keyinfo_h) + siglen;
     maxdest = blocksize / sizeof(struct destkey);
     packetcnt = 1;
     for (i = 0, dests = 0; i < destcount; i++) {
         if (destlist[i].status == DEST_REGISTERED) {
             if (!iv_init) {
                 ivctr++;
-                keyinfo->iv_ctr_hi =htonl((ivctr & 0xFFFFFFFF00000000LL) >> 32);
-                keyinfo->iv_ctr_lo = htonl(ivctr & 0x00000000FFFFFFFFLL);
+                keyinfo->iv_ctr_hi = htonl((ivctr & 0xFFFFFFFF00000000ULL)>>32);
+                keyinfo->iv_ctr_lo = htonl(ivctr & 0x00000000FFFFFFFFULL);
                 iv_init = 1;
             }
             keylist[dests].dest_id = destlist[i].id;
-            build_iv(iv, destlist[i].encinfo->salt, unauth_ivlen,
-                     uftp_htonll(ivctr), header->src_id);
-            if (!encrypt_block(unauth_keytype, iv,destlist[i].encinfo->key,
-                               NULL,0, &groupmaster[1], sizeof(groupmaster) - 1,
-                               keylist[dests].groupmaster, &len)) {
+            build_iv(iv, destlist[i].encinfo->s_hs_iv, ivlen,
+                     uftp_htonll(ivctr));
+            memcpy(aadcontext + baseaadlen, destlist[i].encinfo->context2,
+                    destlist[i].encinfo->context2_len);
+            aadlen = baseaadlen + destlist[i].encinfo->context2_len;
+            if (!encrypt_block(keytype, iv, destlist[i].encinfo->s_hs_key,
+                    aadcontext, aadlen, groupmaster, sizeof(groupmaster),
+                    keylist[dests].groupmaster, &enclen)) {
                 glog0(finfo, "Error encrypting KEYINFO for %s",
                              destlist[i].name);
-                free(buf);
-                free(iv);
-                return 0;
+                rval = 0;
+                goto cleanup;
             }
             dests++;
         }
         if ((dests >= maxdest) || ((i == destcount - 1) && (dests > 0))) {
             header->seq = htons(send_seq++);
-            payloadlen = hsize + (dests * sizeof(struct destkey));
+            packetlen = sizeof(struct uftp_h) + hsize + 
+                        (dests * sizeof(struct destkey));
+
+            memset(sig, 0, siglen);
+            memcpy(sigcontext + basesigcontextlen, buf, packetlen);
+            sigcontextlen = basesigcontextlen + packetlen;
+            if (privkeytype == KEYBLOB_EC) {
+                if (!create_ECDSA_sig(privkey.ec, hashtype, sigcontext, 
+                                      sigcontextlen, sigcopy, &_siglen)) {
+                    // Called function should log
+                    rval = 0;
+                    goto cleanup;
+                }
+            } else {
+                if (!create_RSA_sig(privkey.rsa, hashtype, sigcontext,
+                                    sigcontextlen, sigcopy, &_siglen)) {
+                    // Called function should log
+                    rval = 0;
+                    goto cleanup;
+                }
+            }
+            if (_siglen != siglen) {
+                glog0(finfo, "Signature length doesn't match expected length");
+                glog1(finfo, "expected %d, got %d", siglen, _siglen);
+                rval = 0;
+                goto cleanup;
+            }
+            memcpy(sig, sigcopy, siglen);
+
             glog2(finfo, "Sending KEYINFO %d.%d", attempt, packetcnt);
-            if (nb_sendto(sock, buf, payloadlen + sizeof(struct uftp_h), 0,
+            if (nb_sendto(sock, buf, packetlen, 0,
                           (struct sockaddr *)&receive_dest,
                           family_len(receive_dest)) == SOCKET_ERROR) {
                 gsockerror(finfo, "Error sending KEYINFO");
                 sleep(1);
-                free(buf);
-                free(iv);
-                return 0;
+                rval = 0;
+                goto cleanup;
             }
             if (packet_wait) usleep(packet_wait/1000);
             memset(keylist, 0, maxdest * sizeof(struct destkey));
@@ -316,9 +372,15 @@ int send_keyinfo(const struct finfo_t *finfo, int attempt)
             packetcnt++;
         }
     }
+    rval = 1;
+
+cleanup:
+    free(sigcontext);
+    free(aadcontext);
+    free(sigcopy);
     free(buf);
     free(iv);
-    return 1;
+    return rval;
 }
 
 /**
@@ -333,6 +395,7 @@ int send_fileinfo(const struct finfo_t *finfo, int attempt)
     struct fileinfo_h *fileinfo;
     struct timeval tv;
     uint32_t *idlist;
+    int64_t time_us;
     char *filename, *linkname;
 
     if (strlen(finfo->destfname) > MAXPATHNAME) {
@@ -353,10 +416,13 @@ int send_fileinfo(const struct finfo_t *finfo, int attempt)
     fileinfo->namelen = (uint8_t)(0 + ceil(strlen(finfo->destfname) / 4.0));
     fileinfo->lofsize = htonl((finfo->size & 0xFFFFFFFF));
     fileinfo->hifsize = htons((uint16_t)(finfo->size >> 32));
-    fileinfo->ftstamp = htonl(finfo->tstamp);
+    fileinfo->ftstamp_hi =
+        htons((uint16_t)((finfo->tstamp & 0xFFFFFFFF00000000ULL) >> 32));
+    fileinfo->ftstamp_lo = htonl(finfo->tstamp & 0x00000000FFFFFFFFULL);
     gettimeofday(&tv, NULL);
-    fileinfo->tstamp_sec = htonl(tv.tv_sec);
-    fileinfo->tstamp_usec = htonl(tv.tv_usec);
+    time_us = tv_to_usec(tv);
+    fileinfo->tstamp_hi = htonl((time_us & 0xFFFFFFFF00000000ULL) >> 32);
+    fileinfo->tstamp_lo = htonl(time_us & 0x00000000FFFFFFFFULL);
 
     strncpy(filename, finfo->destfname, MAXPATHNAME);
     if (finfo->ftype == FTYPE_LINK) {
@@ -441,153 +507,66 @@ void add_proxy_dests(struct finfo_t *finfo, const uint32_t *idlist,
     }
 }
 
-/**
- * Returns the verify_data string used in certain messages.  This value
- * is then run through the PRF with the result going into the message
- */
-uint8_t *build_verify_data(const struct finfo_t *finfo, int hostidx,
-                           int *verifylen)
-{
-    uint8_t *verifydata;
-    uint8_t privatemcast[16];
-    uint32_t n_group_id;
-    int iplen;
-
-    if (listen_dest.ss.ss_family == AF_INET6) {
-        iplen = sizeof(struct in6_addr);
-    } else {
-        iplen = sizeof(struct in_addr);
-    }
-    if (!is_multicast(&listen_dest, 0)) {
-        memset(privatemcast, 0, iplen);
-    } else if (listen_dest.ss.ss_family == AF_INET6) {
-        memcpy(privatemcast, &receive_dest.sin6.sin6_addr.s6_addr, iplen);
-    } else {
-        memcpy(privatemcast, &receive_dest.sin.sin_addr.s_addr, iplen);
-    }
-    *verifylen = 0;
-    if (destlist[hostidx].status == DEST_MUTE) {
-        verifydata = safe_calloc(sizeof(finfo->group_id) + iplen +
-                sizeof(rand1) + sizeof(destlist[hostidx].encinfo->rand2) +
-                sizeof(destlist[hostidx].encinfo->premaster), 1);
-    } else {
-        verifydata = safe_calloc(sizeof(finfo->group_id) + iplen +
-                sizeof(rand1) + sizeof(destlist[hostidx].encinfo->rand2) +
-                sizeof(destlist[hostidx].encinfo->premaster) + PUBKEY_LEN +
-                sizeof(groupmaster), 1);
-    }
-
-    n_group_id = htonl(finfo->group_id);
-    memcpy(verifydata, &n_group_id, sizeof(n_group_id));
-    *verifylen += sizeof(n_group_id);
-    memcpy(verifydata + *verifylen, &privatemcast, iplen);
-    *verifylen += iplen;
-    memcpy(verifydata + *verifylen, rand1, sizeof(rand1));
-    *verifylen += sizeof(rand1);
-    memcpy(verifydata + *verifylen, destlist[hostidx].encinfo->rand2,
-            sizeof(destlist[hostidx].encinfo->rand2));
-    *verifylen += sizeof(destlist[hostidx].encinfo->rand2);
-    memcpy(verifydata + *verifylen, destlist[hostidx].encinfo->premaster,
-            destlist[hostidx].encinfo->premaster_len);
-    *verifylen += destlist[hostidx].encinfo->premaster_len;
-
-    if (destlist[hostidx].status != DEST_MUTE) {
-        if (destlist[hostidx].encinfo->pubkey.key) {
-            uint16_t bloblen;
-            uint8_t *keyblob = verifydata + *verifylen;
-
-            if (destlist[hostidx].encinfo->pubkeytype == KEYBLOB_RSA) {
-                if (!export_RSA_key(destlist[hostidx].encinfo->pubkey.rsa,
-                                    keyblob, &bloblen)) {
-                    free(verifydata);
-                    return NULL;
-                }
-            } else {
-                if (!export_EC_key(destlist[hostidx].encinfo->pubkey.ec,
-                                   keyblob, &bloblen)) {
-                    glog0(finfo, "Error exporting server public key");
-                    free(verifydata);
-                    return NULL;
-                }
-            }
-            *verifylen += bloblen;
-        }
-        memcpy(verifydata + *verifylen, groupmaster, sizeof(groupmaster));
-        *verifylen += sizeof(groupmaster);
-    }
-
-    return verifydata;
-}
-
 /**
  * Verifies the data in a CLIENT_KEY message signed by the client's public key
  */
-int verify_client_key(struct finfo_t *finfo, int hostidx)
+int verify_client_key(struct finfo_t *finfo, int hostidx, 
+                      uint8_t *sig, struct client_key_h *clientkey)
 {
     uint8_t *verifydata;
     int verifylen;
+    const char *context_str = "UFTP 5, CLIENT_KEY";
+    struct encinfo_t *encinfo = destlist[hostidx].encinfo;
 
-    // build_verify_data should never fail in this case
-    verifydata = build_verify_data(finfo, hostidx, &verifylen);
+    verifydata = safe_calloc(strlen(context_str) + 
+            encinfo->context1_len, 1);
+    verifylen = 0;
+    memcpy(verifydata + verifylen, context_str, strlen(context_str));
+    verifylen += (int)strlen(context_str);
+    memcpy(verifydata + verifylen, encinfo->context1,
+            encinfo->context1_len);
+    verifylen += encinfo->context1_len;
 
-    if (destlist[hostidx].encinfo->pubkeytype == KEYBLOB_RSA) {
-        if (!verify_RSA_sig(destlist[hostidx].encinfo->pubkey.rsa, hashtype,
-                verifydata, verifylen, destlist[hostidx].encinfo->verifydata,
-                destlist[hostidx].encinfo->verifylen)) {
+    if (encinfo->pubkeytype == KEYBLOB_RSA) {
+        if (!verify_RSA_sig(encinfo->pubkey.rsa, hashtype, verifydata,
+                verifylen, sig, ntohs(clientkey->siglen))) {
             glog1(finfo, "Rejecting CLIENT_KEY from %s: verify data mismatch",
                          destlist[hostidx].name);
-            free(verifydata);
-            return 0;
+            goto err_exit;
         }
     } else {
-        if (!verify_ECDSA_sig(destlist[hostidx].encinfo->pubkey.ec, hashtype,
-                verifydata, verifylen, destlist[hostidx].encinfo->verifydata,
-                destlist[hostidx].encinfo->verifylen)) {
+        if (!verify_ECDSA_sig(encinfo->pubkey.ec, hashtype, verifydata,
+                              verifylen, sig, ntohs(clientkey->siglen))) {
             glog1(finfo, "Rejecting CLIENT_KEY from %s: verify data mismatch",
                          destlist[hostidx].name);
-            free(verifydata);
-            return 0;
+            goto err_exit;
         }
     }
 
+    // sig should point to the signature field in clientkey
+    memset(sig, 0, ntohs(clientkey->siglen));
+    create_client_context_2(encinfo->context1, encinfo->context1_len,
+                            clientkey, clientkey->hlen * 4,
+                            &encinfo->context2, &encinfo->context2_len);
+    calculate_client_app_keys(hashtype, groupmaster, sizeof(groupmaster),
+            encinfo->context2, encinfo->context2_len, keylen, ivlen, 
+            encinfo->c_app_key, encinfo->c_app_iv,
+            encinfo->finished_key, encinfo->verify_data);
+
     destlist[hostidx].status = DEST_REGISTERED;
     free(verifydata);
     return 1;
-}
 
-/**
- * For a given client, calculate the master key and do key expansion
- * to determine the symmetric cypher key and IV salt, and hash key
- */
-void calculate_client_keys(int hostidx)
-{
-    unsigned char *seed, *prf_buf;
-    int explen, len, seedlen;
-
-    explen = keylen + SALT_LEN + hmaclen;
-    seedlen = sizeof(rand1) * 2;
-    seed = safe_calloc(seedlen, 1);
-    prf_buf = safe_calloc(MASTER_LEN + explen + hmaclen, 1);
-
-    memcpy(seed, rand1, sizeof(rand1));
-    memcpy(seed + sizeof(rand1), destlist[hostidx].encinfo->rand2,
-            sizeof(destlist[hostidx].encinfo->rand2));
-    PRF(hashtype, MASTER_LEN, destlist[hostidx].encinfo->premaster,
-            destlist[hostidx].encinfo->premaster_len,
-            "master secret", seed, seedlen, prf_buf, &len);
-    memcpy(destlist[hostidx].encinfo->master,prf_buf,
-            sizeof(destlist[hostidx].encinfo->master));
-
-    PRF(hashtype, explen, destlist[hostidx].encinfo->master, 
-            sizeof(destlist[hostidx].encinfo->master), "key expansion",
-            seed, seedlen, prf_buf, &len);
-    memcpy(destlist[hostidx].encinfo->hmackey, prf_buf, hmaclen);
-    memcpy(destlist[hostidx].encinfo->key, prf_buf + hmaclen, keylen);
-    memcpy(destlist[hostidx].encinfo->salt, prf_buf + hmaclen + keylen,
-            SALT_LEN);
-
-    free(seed);
-    free(prf_buf);
+err_exit:
+    if (destlist[hostidx].encinfo->pubkeytype == KEYBLOB_RSA) {
+        free_RSA_key(destlist[hostidx].encinfo->pubkey.rsa);
+    } else {
+        free_EC_key(destlist[hostidx].encinfo->pubkey.ec);
+    }
+    destlist[hostidx].encinfo->pubkey.key = 0;
+    destlist[hostidx].encinfo->pubkeytype = 0;
+    free(verifydata);
+    return 0;
 }
 
 /**
@@ -597,53 +576,48 @@ int handle_register_keys(const struct register_h *reg,
                          const unsigned char *keyinfo, struct finfo_t *finfo,
                          int hostidx)
 {
-    unsigned char premaster[PUBKEY_LEN];
-    unsigned int len;
-
-    destlist[hostidx].encinfo = safe_calloc(1, sizeof(struct encinfo_t));
-    memcpy(destlist[hostidx].encinfo->rand2, reg->rand2,
-           sizeof(destlist[hostidx].encinfo->rand2));
-    if (keyextype == KEYEX_RSA) {
-        if (!RSA_decrypt(privkey.rsa, keyinfo, ntohs(reg->keyinfo_len),
-                         premaster, &len)) {
-            glog1(finfo, "Rejecting REGISTER from %s: failed to decrypt "
-                         "premaster secret", destlist[hostidx].name);
-            return 0;
-        }
-        if (len != MASTER_LEN) {
-            glog1(finfo, "Rejecting REGISTER from %s: decrypted premaster "
-                         "secret wrong length", destlist[hostidx].name);
-            return 0;
-        }
-    } else {
-        if (!import_EC_key(&destlist[hostidx].encinfo->dhkey.ec, keyinfo,
-                           ntohs(reg->keyinfo_len), 1)) {
-            glog1(finfo, "Rejecting REGISTER from %s: "
-                         "failed to import ECDH key", destlist[hostidx].name);
-            return 0;
-        }
-        if (get_EC_curve(destlist[hostidx].encinfo->dhkey.ec) != ecdh_curve) {
-            glog1(finfo, "Rejecting REGISTER from %s: "
-                         "invalid curve for ECDH", destlist[hostidx].name);
-            return 0;
-        }
-        if (!get_ECDH_key(destlist[hostidx].encinfo->dhkey.ec, dhkey.ec,
-                          premaster, &len)) {
-            glog1(finfo, "Rejecting REGISTER from %s: failed to calculate "
-                         "premaster secret", destlist[hostidx].name);
-            return 0;
-        }
+    struct encinfo_t *encinfo = safe_calloc(1, sizeof(struct encinfo_t));
+    memcpy(encinfo->rand2, reg->rand2, sizeof(encinfo->rand2));
+    if (!import_EC_key(&encinfo->dhkey.ec, keyinfo,
+                       ntohs(reg->dhlen), 1)) {
+        glog1(finfo, "Rejecting REGISTER from %s: failed to import ECDH key",
+                     destlist[hostidx].name);
+        free(encinfo);
+        return 0;
     }
-    memcpy(destlist[hostidx].encinfo->premaster, premaster, len);
-    destlist[hostidx].encinfo->premaster_len = len;
-    calculate_client_keys(hostidx);
-
-    if (destlist[hostidx].encinfo->pubkey.key) {
-        if (!verify_client_key(finfo, hostidx)) {
-            return 0;
-        }
+    if (get_EC_curve(encinfo->dhkey.ec) != ecdh_curve) {
+        glog1(finfo, "Rejecting REGISTER from %s: " "invalid curve for ECDH",
+                     destlist[hostidx].name);
+        free_EC_key(encinfo->dhkey.ec);
+        free(encinfo);
+        return 0;
+    }
+    if (!get_ECDH_key(encinfo->dhkey.ec, dhkey.ec, encinfo->premaster,
+                      &encinfo->premaster_len, HASH_SHA256)) {
+        glog1(finfo, "Rejecting REGISTER from %s: failed to calculate "
+                     "premaster secret", destlist[hostidx].name);
+        free_EC_key(encinfo->dhkey.ec);
+        free(encinfo);
+        return 0;
     }
 
+    create_client_context_1(servercontext, servercontext_len, NULL, 0,
+            destlist[hostidx].id, keyinfo, ntohs(reg->dhlen), reg->rand2,
+            &encinfo->context1, &encinfo->context1_len);
+    calculate_hs_keys(hashtype, encinfo->premaster, encinfo->premaster_len,
+                      encinfo->context1, encinfo->context1_len, keylen, ivlen,
+                      encinfo->s_hs_key, encinfo->s_hs_iv,
+                      encinfo->c_hs_key, encinfo->c_hs_iv);
+    if (!client_auth) {
+        create_client_context_2(encinfo->context1, encinfo->context1_len,
+                NULL, 0, &encinfo->context2, &encinfo->context2_len);
+        calculate_client_app_keys(hashtype, groupmaster, sizeof(groupmaster),
+                encinfo->context2, encinfo->context2_len, keylen, ivlen, 
+                encinfo->c_app_key, encinfo->c_app_iv,
+                encinfo->finished_key, encinfo->verify_data);
+    }
+
+    destlist[hostidx].encinfo = encinfo;
     return 1;
 }
 
@@ -658,11 +632,13 @@ void handle_open_register(const unsigned char *message, unsigned meslen,
     const uint32_t *idlist;
     const unsigned char *enckey;
     int clientcnt, hostidx;
-    struct timeval tv1, tv2;
+    struct timeval tv;
+    int64_t t1, t2;
 
     reg = (const struct register_h *)message;
     enckey = (const unsigned char *)reg + sizeof(struct register_h);
-    gettimeofday(&tv2, NULL);
+    gettimeofday(&tv, NULL);
+    t2 = tv_to_usec(tv);
 
     if (destcount == MAXDEST) {
         glog1(finfo, "Rejecting REGISTER from %08X: "
@@ -671,7 +647,7 @@ void handle_open_register(const unsigned char *message, unsigned meslen,
         return;
     }
     if ((meslen < (reg->hlen * 4U)) || ((reg->hlen * 4U) <
-            sizeof(struct register_h) + ntohs(reg->keyinfo_len))) {
+            sizeof(struct register_h) + ntohs(reg->dhlen))) {
         glog1(finfo, "Rejecting REGISTER from %08X: "
                      "invalid message size", ntohl(src));
         send_abort(finfo, "Invalid message size", su, src, 0, 0);
@@ -688,9 +664,9 @@ void handle_open_register(const unsigned char *message, unsigned meslen,
     if (regconf) {
         finfo->deststate[hostidx].conf_sent = 0;
     }
-    tv1.tv_sec = ntohl(reg->tstamp_sec);
-    tv1.tv_usec = ntohl(reg->tstamp_usec);
-    destlist[hostidx].rtt = diff_usec(tv2, tv1) / 1000000.0;
+    t1 = (int64_t)ntohl(reg->tstamp_hi) << 32;
+    t1 |= ntohl(reg->tstamp_lo);
+    destlist[hostidx].rtt = (t2 - t1) / 1000000.0;
     if (destlist[hostidx].rtt < CLIENT_RTT_MIN) {
         destlist[hostidx].rtt = CLIENT_RTT_MIN;
     }
@@ -705,14 +681,13 @@ void handle_open_register(const unsigned char *message, unsigned meslen,
         add_proxy_dests(finfo, idlist, su, clientcnt, hostidx, 1,
                         destlist[hostidx].rtt);
     }
-    glog3(finfo, "send time = %d.%06d", tv1.tv_sec, tv1.tv_usec);
-    glog3(finfo, "rx time = %d.%06d", tv2.tv_sec, tv2.tv_usec);
+    glog3(finfo, "send time = " F_i64 ".%06d", t1/1000000, (int)(t1%1000000));
+    glog3(finfo, "rx time = " F_i64 ".%06d", t2/1000000, (int)(t2%1000000));
     glog3(finfo, "  rtt = %.6f", destlist[hostidx].rtt);
 }
 
 /**
- * Process an expected REGISTER with closed group membership,
- * or with open group membership if CLIENT_KEY was received first.
+ * Process an expected REGISTER with closed group membership
  */
 void handle_register(const unsigned char *message, unsigned meslen,
                      struct finfo_t *finfo, const union sockaddr_u *su,
@@ -722,14 +697,16 @@ void handle_register(const unsigned char *message, unsigned meslen,
     const uint32_t *idlist;
     const unsigned char *enckey;
     int clientcnt, dupmsg, isproxy;
-    struct timeval tv1, tv2;
+    struct timeval tv;
+    int64_t t1, t2;
 
     reg = (const struct register_h *)message;
     enckey = (const unsigned char *)reg + sizeof(struct register_h);
-    gettimeofday(&tv2, NULL);
+    gettimeofday(&tv, NULL);
+    t2 = tv_to_usec(tv);
 
     if ((meslen < (reg->hlen * 4U)) || ((reg->hlen * 4U) <
-            sizeof(struct register_h) + ntohs(reg->keyinfo_len))) {
+            sizeof(struct register_h) + ntohs(reg->dhlen))) {
         glog1(finfo, "Rejecting REGISTER from %s: "
                      "invalid message size", destlist[hostidx].name);
         send_abort(finfo, "Invalid message size", su, destlist[hostidx].id,0,0);
@@ -750,7 +727,7 @@ void handle_register(const unsigned char *message, unsigned meslen,
     }
 
     if (destlist[hostidx].status == DEST_MUTE) {
-        if (keytype != KEY_NONE) {
+        if ((keytype != KEY_NONE) && !destlist[hostidx].encinfo) {
             if (!handle_register_keys(reg, enckey, finfo, hostidx)) {
                 return;
             }
@@ -760,9 +737,9 @@ void handle_register(const unsigned char *message, unsigned meslen,
                     ? DEST_MUTE : DEST_REGISTERED);
     }
     dupmsg = (destlist[hostidx].registered);
-    tv1.tv_sec = ntohl(reg->tstamp_sec);
-    tv1.tv_usec = ntohl(reg->tstamp_usec);
-    destlist[hostidx].rtt = diff_usec(tv2, tv1) / 1000000.0;
+    t1 = (int64_t)ntohl(reg->tstamp_hi) << 32;
+    t1 |= ntohl(reg->tstamp_lo);
+    destlist[hostidx].rtt = (t2 - t1) / 1000000.0;
     if (destlist[hostidx].rtt < CLIENT_RTT_MIN) {
         destlist[hostidx].rtt = CLIENT_RTT_MIN;
     }
@@ -780,8 +757,8 @@ void handle_register(const unsigned char *message, unsigned meslen,
         add_proxy_dests(finfo, idlist, su, clientcnt, hostidx, open,
                         destlist[hostidx].rtt);
     }
-    glog3(finfo, "send time = %d.%06d", tv1.tv_sec, tv1.tv_usec);
-    glog3(finfo, "rx time = %d.%06d", tv2.tv_sec, tv2.tv_usec);
+    glog3(finfo, "send time = " F_i64 ".%06d", t1/1000000, (int)(t1%1000000));
+    glog3(finfo, "rx time = " F_i64 ".%06d", t2/1000000, (int)(t2%1000000));
     glog3(finfo, "  rtt = %.6f", destlist[hostidx].rtt);
 }
 
@@ -795,29 +772,24 @@ int verify_client_fingerprint(const struct finfo_t *finfo,
     unsigned char fingerprint[HMAC_LEN];
     unsigned int fplen;
 
-    if (destlist[hostidx].verified) {
-        return 1;
-    }
-
-    destlist[hostidx].encinfo->pubkeytype = keyblob[0];
     if (keyblob[0] == KEYBLOB_RSA) {
         if (!import_RSA_key(&destlist[hostidx].encinfo->pubkey.rsa,
                             keyblob, bloblen)) {
             glog1(finfo, "Rejecting CLIENT_KEY from %s: "
                          "failed to import key", destlist[hostidx].name);
+            destlist[hostidx].encinfo->pubkey.key = 0;
             return 0;
         }
-        destlist[hostidx].encinfo->pubkeylen =
-                RSA_keylen(destlist[hostidx].encinfo->pubkey.rsa);
+        destlist[hostidx].encinfo->pubkeytype = KEYBLOB_RSA;
     } else {
         if (!import_EC_key(&destlist[hostidx].encinfo->pubkey.ec,
                            keyblob, bloblen, 0)) {
             glog1(finfo, "Rejecting CLIENT_KEY from %s: "
                          "failed to import key", destlist[hostidx].name);
+            destlist[hostidx].encinfo->pubkey.key = 0;
             return 0;
         }
-        destlist[hostidx].encinfo->pubkeylen =
-                ECDSA_siglen(destlist[hostidx].encinfo->pubkey.ec);
+        destlist[hostidx].encinfo->pubkeytype = KEYBLOB_EC;
     }
 
     if (destlist[hostidx].has_fingerprint) {
@@ -831,79 +803,26 @@ int verify_client_fingerprint(const struct finfo_t *finfo,
                 free_EC_key(destlist[hostidx].encinfo->pubkey.ec);
             }
             destlist[hostidx].encinfo->pubkey.key = 0;
-            destlist[hostidx].encinfo->pubkeylen = 0;
+            destlist[hostidx].encinfo->pubkeytype = 0;
             return 0;
         }
     }
 
-    destlist[hostidx].verified = 1;
     return 1;
 }
 
 /**
- * Process an expected CLIENT_KEY with open group membership
+ * Process an expected CLIENT_KEY
  */
-void handle_open_clientkey(const unsigned char *message, unsigned meslen,
-                           struct finfo_t *finfo, const union sockaddr_u *su,
-                           uint32_t src)
-{
-    const struct client_key_h *clientkey;
-    const unsigned char *keyblob, *verify;
-    int hostidx;
-
-    clientkey = (const struct client_key_h *)message;
-    keyblob = (const unsigned char *)clientkey + sizeof(struct client_key_h);
-    verify = keyblob + ntohs(clientkey->bloblen);
-
-    if (destcount == MAXDEST) {
-        glog1(finfo, "Rejecting CLIENT_KEY from %08X: "
-                     "max destinations exceeded", ntohl(src));
-        send_abort(finfo, "Max destinations exceeded", su, src, 0, 0);
-        return;
-    }
-    if ((meslen < (clientkey->hlen * 4U)) ||
-            ((clientkey->hlen * 4U) < sizeof(struct client_key_h) +
-                ntohs(clientkey->bloblen) + ntohs(clientkey->siglen))) {
-        glog1(finfo, "Rejecting CLIENT_KEY from %08X: "
-                     "invalid message size", ntohl(src));
-        send_abort(finfo, "Invalid message size", su, src, 0, 0);
-        return;
-    }
-    if (sigtype == SIG_KEYEX) {
-        if ((((keyextype == KEYEX_RSA) || (keyextype == KEYEX_ECDH_RSA)) &&
-                    (keyblob[0] != KEYBLOB_RSA)) ||
-                ((keyextype == KEYEX_ECDH_ECDSA) &&
-                 (keyblob[0] != KEYBLOB_EC))) {
-            glog1(finfo, "Rejecting CLIENT_KEY from %08X: "
-                         "invalid keyblob type", ntohl(src));
-            return;
-        }
-    }
-
-    hostidx = add_dest_by_addr(src, finfo, DEST_MUTE, -1, 0);
-    if (!verify_client_fingerprint(finfo, keyblob, ntohs(clientkey->bloblen),
-                                   hostidx)) {
-        return;
-    }
-    memcpy(destlist[hostidx].encinfo->verifydata, verify,
-            ntohs(clientkey->siglen));
-    destlist[hostidx].encinfo->verifylen = ntohs(clientkey->siglen);
-    glog2(finfo, "Received CLIENT_KEY from %s", destlist[hostidx].name);
-}
-
-/**
- * Process an expected CLIENT_KEY with closed group membership,
- * or with open group membership if REGISTER was received first.
- */
-void handle_clientkey(const unsigned char *message, unsigned meslen,
+void handle_clientkey(unsigned char *message, unsigned meslen,
                       struct finfo_t *finfo, const union sockaddr_u *su,
                       int hostidx)
 {
-    const struct client_key_h *clientkey;
-    const unsigned char *keyblob, *verify;
+    struct client_key_h *clientkey;
+    unsigned char *keyblob, *verify;
 
-    clientkey = (const struct client_key_h *)message;
-    keyblob = (const unsigned char *)clientkey + sizeof(struct client_key_h);
+    clientkey = (struct client_key_h *)message;
+    keyblob = (unsigned char *)clientkey + sizeof(struct client_key_h);
     verify = keyblob + ntohs(clientkey->bloblen);
 
     if ((meslen < (clientkey->hlen * 4U)) ||
@@ -914,16 +833,6 @@ void handle_clientkey(const unsigned char *message, unsigned meslen,
         send_abort(finfo, "Invalid message size", su, destlist[hostidx].id,0,0);
         return;
     }
-    if (sigtype == SIG_KEYEX) {
-        if ((((keyextype == KEYEX_RSA) || (keyextype == KEYEX_ECDH_RSA)) &&
-                    (keyblob[0] != KEYBLOB_RSA)) ||
-                ((keyextype == KEYEX_ECDH_ECDSA) &&
-                 (keyblob[0] != KEYBLOB_EC))) {
-            glog1(finfo, "Rejecting CLIENT_KEY from %s: "
-                         "invalid keyblob type", destlist[hostidx].name);
-            return;
-        }
-    }
     if (finfo->file_id != 0) {
         glog2(finfo, "Received CLIENT_KEY+ from %s", destlist[hostidx].name);
         return;
@@ -933,13 +842,8 @@ void handle_clientkey(const unsigned char *message, unsigned meslen,
                                    hostidx)) {
         return;
     }
-    memcpy(destlist[hostidx].encinfo->verifydata, verify,
-            ntohs(clientkey->siglen));
-    destlist[hostidx].encinfo->verifylen = ntohs(clientkey->siglen);
-    if (destlist[hostidx].registered) {
-        if (!verify_client_key(finfo, hostidx)) {
-            return;
-        }
+    if (!verify_client_key(finfo, hostidx, verify, clientkey)) {
+        return;
     }
     glog2(finfo, "Received CLIENT_KEY from %s", destlist[hostidx].name);
 }
@@ -952,14 +856,14 @@ void handle_keyinfo_ack(const unsigned char *message, unsigned meslen,
                         int hostidx)
 {
     const struct keyinfoack_h *keyinfoack;
-    unsigned char *verifydata, *verify_hash, *verify_test;
-    int verifylen, len, dupmsg;
-    unsigned int hashlen;
+    const uint8_t *verify_data;
+    int dupmsg;
 
     keyinfoack = (const struct keyinfoack_h *)message;
+    verify_data = (const uint8_t *)message + sizeof(struct keyinfoack_h);
 
     if ((meslen < (keyinfoack->hlen * 4U)) ||
-            ((keyinfoack->hlen * 4U) < sizeof(struct keyinfoack_h))) {
+           ((keyinfoack->hlen * 4U) < (sizeof(struct keyinfoack_h) + hashlen))){
         glog1(finfo, "Rejecting KEYINFO_ACK from %s: "
                      "invalid message size", destlist[hostidx].name);
         send_abort(finfo, "Invalid message size", su, destlist[hostidx].id,0,0);
@@ -974,29 +878,12 @@ void handle_keyinfo_ack(const unsigned char *message, unsigned meslen,
         return;
     }
 
-    if (!(verifydata = build_verify_data(finfo, hostidx, &verifylen))) {
-        glog1(finfo, "Rejecting KEYINFO_ACK from %s: "
-                     "error exporting client public key", destlist[hostidx].name);
-        return;
-    }
-    verify_hash = safe_calloc(hmaclen, 1);
-    verify_test = safe_calloc(VERIFY_LEN + hmaclen, 1);
-    hash(hashtype, verifydata, verifylen, verify_hash, &hashlen);
-    PRF(hashtype, VERIFY_LEN, groupmaster, sizeof(groupmaster),
-            "client finished", verify_hash, hashlen, verify_test, &len);
-    if (memcmp(keyinfoack->verify_data, verify_test, VERIFY_LEN)) {
+    if (memcmp(verify_data, destlist[hostidx].encinfo->verify_data, hashlen)) {
         glog1(finfo, "Rejecting KEYINFO_ACK from %s: "
                      "verify data mismatch", destlist[hostidx].name);
-        free(verifydata);
-        free(verify_hash);
-        free(verify_test);
         return;
     }
 
-    free(verifydata);
-    free(verify_hash);
-    free(verify_test);
-
     dupmsg = (destlist[hostidx].status == DEST_ACTIVE);
     glog2(finfo, "Received KEYINFO_ACK%s from %s", dupmsg ? "+" : "",
                  destlist[hostidx].name);
@@ -1010,12 +897,14 @@ void handle_fileinfo_ack(const unsigned char *message, unsigned meslen,
                          struct finfo_t *finfo, int hostidx)
 {
     const struct fileinfoack_h *fileinfoack;
-    struct timeval tv1, tv2;
+    struct timeval tv;
+    int64_t t1, t2;
     const uint32_t *addr;
     int clientcnt, dupmsg, isproxy, clientidx, i;
 
     fileinfoack = (const struct fileinfoack_h *)message;
-    gettimeofday(&tv2, NULL);
+    gettimeofday(&tv, NULL);
+    t2 = tv_to_usec(tv);
 
     if ((meslen < (fileinfoack->hlen * 4U)) ||
             ((fileinfoack->hlen * 4U) < sizeof(struct fileinfoack_h))) {
@@ -1040,9 +929,9 @@ void handle_fileinfo_ack(const unsigned char *message, unsigned meslen,
     finfo->partial = finfo->partial &&
             ((fileinfoack->flags & FLAG_PARTIAL) != 0);
 
-    tv1.tv_sec = ntohl(fileinfoack->tstamp_sec);
-    tv1.tv_usec = ntohl(fileinfoack->tstamp_usec);
-    destlist[hostidx].rtt = diff_usec(tv2, tv1) / 1000000.0;
+    t1 = (int64_t)ntohl(fileinfoack->tstamp_hi) << 32;
+    t1 |= ntohl(fileinfoack->tstamp_lo);
+    destlist[hostidx].rtt = (t2 - t1) / 1000000.0;
     if (destlist[hostidx].rtt < CLIENT_RTT_MIN) {
         destlist[hostidx].rtt = CLIENT_RTT_MIN;
     }
@@ -1071,9 +960,7 @@ void handle_fileinfo_ack(const unsigned char *message, unsigned meslen,
                          destlist[clientidx].name);
         }
     }
-    glog3(finfo, "send time = %d.%06d", tv1.tv_sec, tv1.tv_usec);
-    glog3(finfo, "rx time = %d.%06d", tv2.tv_sec, tv2.tv_usec);
+    glog3(finfo, "send time = " F_i64 ".%06d", t1/1000000, (int)(t1%1000000));
+    glog3(finfo, "rx time = " F_i64 ".%06d", t2/1000000, (int)(t2%1000000));
     glog3(finfo, "  rtt = %.6f", destlist[hostidx].rtt);
-
-    return;
 }
