@@ -66,8 +66,8 @@ void cleanup(void)
     int i;
 
     closesocket(sock);
-    for (i = 0; i < destcount; i++) {
-        if (keytype != KEY_NONE) {
+    if (keytype != KEY_NONE) {
+        for (i = 0; i < destcount; i++) {
             if (destlist[i].encinfo) {
                 if (destlist[i].encinfo->pubkey.key) {
                     if (destlist[i].encinfo->pubkeytype == KEYBLOB_RSA) {
@@ -76,23 +76,19 @@ void cleanup(void)
                         free_EC_key(destlist[i].encinfo->pubkey.ec);
                     }
                 }
-                if ((keyextype == KEYEX_ECDH_RSA) ||
-                        (keyextype == KEYEX_ECDH_ECDSA)) {
-                    free_EC_key(destlist[i].encinfo->dhkey.ec);
-                }
+                free_EC_key(destlist[i].encinfo->dhkey.ec);
+                free(destlist[i].encinfo->context1);
+                free(destlist[i].encinfo->context2);
                 free(destlist[i].encinfo);
             }
         }
-    }
-    if (keytype != KEY_NONE) {
-        if ((keyextype == KEYEX_RSA) || (keyextype == KEYEX_ECDH_RSA)) {
+        if (privkeytype == KEYBLOB_RSA) {
             free_RSA_key(privkey.rsa);
         } else {
             free_EC_key(privkey.ec);
         }
-        if ((keyextype == KEYEX_ECDH_RSA) || (keyextype == KEYEX_ECDH_ECDSA)) {
-            free_EC_key(dhkey.ec);
-        }
+        free_EC_key(dhkey.ec);
+        free(servercontext);
     }
     crypto_cleanup();
     if (status_file) {
@@ -273,7 +269,8 @@ void create_sockets(void)
         log0(0, 0, 0, "Error getting bind address: %s", gai_strerror(rval));
         exit(ERR_SOCKET);
     }
-    if (bind(sock, ai_rval->ai_addr, ai_rval->ai_addrlen) == SOCKET_ERROR) {
+    if (bind(sock, ai_rval->ai_addr,
+            (socklen_t)ai_rval->ai_addrlen) == SOCKET_ERROR) {
         sockerror(0, 0, 0, "Error binding socket");
         exit(ERR_SOCKET);
     }
@@ -429,76 +426,55 @@ void create_sockets(void)
  */
 void key_init(void)
 {
-    unsigned char *prf_buf;
-    time_t t;
-    uint32_t t2;
-    int explen, len;
-
     if (keytype == KEY_NONE) {
         return;
     }
 
     set_sys_keys(sys_keys);
     get_key_info(keytype, &keylen, &ivlen);
-    hmaclen = get_hash_len(hashtype);
+    hashlen = get_hash_len(hashtype);
 
-    memset(groupkey, 0, sizeof(groupkey));
-    memset(groupsalt, 0, sizeof(groupsalt));
-    memset(grouphmackey, 0, sizeof(grouphmackey));
+    memset(server_app_key, 0, sizeof(server_app_key));
+    memset(server_app_iv, 0, sizeof(server_app_iv));
+    servercontext = NULL;
+    servercontext_len = 0;
 
     if (!get_random_bytes(groupmaster, sizeof(groupmaster))) {
         log0(0, 0, 0, "Failed to generate group master");
         exit(ERR_CRYPTO);
     }
-    groupmaster[0] = UFTP_VER_NUM;
     if (!get_random_bytes(rand1, sizeof(rand1))) {
         log0(0, 0, 0, "Failed to generate rand1");
         exit(ERR_CRYPTO);
     }
-    // Sets the first 4 bytes of rand1 to the current time
-    t = time(NULL);
-    t2 = (uint32_t)(t & 0xFFFFFFFF);
-    memcpy(&rand1, &t2, sizeof(t2));
 
-    explen = hmaclen + keylen + SALT_LEN;
-    prf_buf = safe_calloc(explen + hmaclen, 1);
-    PRF(hashtype, explen, groupmaster, sizeof(groupmaster), "key expansion",
-            rand1, sizeof(rand1), prf_buf, &len);
-    memcpy(grouphmackey, prf_buf, hmaclen);
-    memcpy(groupkey, prf_buf + hmaclen, keylen);
-    memcpy(groupsalt, prf_buf + hmaclen + keylen, SALT_LEN);
-    ivctr = 0;
-    free(prf_buf);
-
-    if ((keyextype == KEYEX_RSA) || (keyextype == KEYEX_ECDH_RSA)) {
-        if ((!strcmp(keyfile, "")) || (newkeylen != 0)) {
-            privkey.rsa = gen_RSA_key(newkeylen, RSA_EXP, keyfile);
-        } else {
-            privkey.rsa = read_RSA_key(keyfile);
+    if (strcmp(keyfile, "") && (newkeylen == 0) && (ecdsa_curve == 0)) {
+        privkey = read_private_key(keyfile, &privkeytype);
+        if (keytype == 0) {
+            log0(0, 0, 0, "Failed to read private key");
+            exit(ERR_CRYPTO);
         }
+    } else if (newkeylen != 0) {
+        privkeytype = KEYBLOB_RSA;
+        privkey.rsa = gen_RSA_key(newkeylen, RSA_EXP, keyfile);
         if (!privkey.key) {
-            log0(0, 0, 0, "Failed to read/generate private key");
+            log0(0, 0, 0, "Failed to generate RSA private key");
             exit(ERR_CRYPTO);
         }
-        privkeylen = RSA_keylen(privkey.rsa);
     } else {
-        if ((!strcmp(keyfile, "")) || (ecdsa_curve != 0)) {
-            privkey.ec = gen_EC_key(ecdsa_curve, 0, keyfile);
-        } else {
-            privkey.ec = read_EC_key(keyfile);
-        }
+        privkeytype = KEYBLOB_EC;
+        privkey.ec = gen_EC_key(ecdsa_curve ? ecdsa_curve : DEF_CURVE,
+                                0, keyfile);
         if (!privkey.key) {
-            log0(0, 0, 0, "Failed to read/generate private key");
+            log0(0, 0, 0, "Failed to generate EC private key");
             exit(ERR_CRYPTO);
         }
-        privkeylen = ECDSA_siglen(privkey.ec);
     }
-    if ((keyextype == KEYEX_ECDH_RSA) || (keyextype == KEYEX_ECDH_ECDSA)) {
-        dhkey.ec = gen_EC_key(ecdh_curve, 1, NULL);
-        if (!dhkey.key) {
-            log0(0, 0, 0, "Failed to generate DH key");
-            exit(ERR_CRYPTO);
-        }
+
+    dhkey.ec = gen_EC_key(ecdh_curve, 1, NULL);
+    if (!dhkey.key) {
+        log0(0, 0, 0, "Failed to generate DH key");
+        exit(ERR_CRYPTO);
     }
 }
 
@@ -547,9 +523,7 @@ void initialize(void)
         datapacketsize += sizeof(struct tfmcc_data_info_he);
     }
     if (keytype != KEY_NONE) {
-        datapacketsize += ((sigtype == SIG_KEYEX) ? privkeylen :
-                (sigtype == SIG_HMAC) ? hmaclen : 0) +
-                KEYBLSIZE + sizeof(struct encrypted_h);
+        datapacketsize += KEYBLSIZE + sizeof(struct encrypted_h);
     }
     // 8 = UDP size, 20 = IPv4 size, 40 = IPv6 size
     if (listen_dest.ss.ss_family == AF_INET6) {
@@ -558,15 +532,12 @@ void initialize(void)
         datapacketsize += sizeof(struct uftp_h) + 8 + 20;
     }
 
-    // Never ask for a client key with no encryption,
-    // and always ask with RSA/ECDSA signatures
+    // Never ask for a client key with no encryption
     if (keytype == KEY_NONE) {
         client_auth = 0;
-    } else if (sigtype == SIG_KEYEX) {
-        client_auth = 1;
     }
 
-    if (cc_type == CC_NONE || cc_type == CC_UFTP3) {
+    if (cc_type == CC_NONE) {
         if (rate == -1) {
             packet_wait = 0;
         } else {
