@@ -68,7 +68,7 @@ int handle_announce_phase(unsigned char *packet, unsigned char *decrypted,
 {
     struct uftp_h *header;
     unsigned char *message;
-    int hostidx;
+    int hostidx, log_sav;
     unsigned decryptlen, meslen;
     uint8_t *func;
 
@@ -76,14 +76,23 @@ int handle_announce_phase(unsigned char *packet, unsigned char *decrypted,
     hostidx = find_client(header->src_id);
     if ((keytype != KEY_NONE) && (header->func == ENCRYPTED)) {
         if (hostidx == -1) {
-            glog2(finfo, "Got encrypted packet from unknown receiver %08X",
+            glog3(finfo, "Got encrypted packet from unknown receiver %08X",
                          ntohl(header->src_id));
             return 0;
         }
-        if (!validate_and_decrypt(packet, packetlen, &decrypted, &decryptlen,
-                keytype, groupkey, groupsalt, ivlen, hashtype, grouphmackey,
-                hmaclen, sigtype, keyextype, destlist[hostidx].encinfo->pubkey,
-                destlist[hostidx].encinfo->pubkeylen)) {
+        // first try c_hs_key, then c_app_key
+        func = NULL;
+        log_sav = log_level;
+        log_level= -1;
+        if (client_auth && validate_and_decrypt(packet, packetlen, &decrypted,
+                    &decryptlen, keytype, destlist[hostidx].encinfo->c_hs_key,
+                    destlist[hostidx].encinfo->c_hs_iv, ivlen)) {
+            func = (uint8_t *)decrypted;
+        }
+        log_level = log_sav;
+        if (!func && !validate_and_decrypt(packet, packetlen, &decrypted, 
+                &decryptlen, keytype, destlist[hostidx].encinfo->c_app_key, 
+                destlist[hostidx].encinfo->c_app_iv, ivlen)) {
             glog1(finfo, "Rejecting message from %s: decrypt/validate failed",
                          destlist[hostidx].name);
             return 0;
@@ -92,7 +101,8 @@ int handle_announce_phase(unsigned char *packet, unsigned char *decrypted,
         message = decrypted;
         meslen = decryptlen;
     } else {
-        if ((keytype != KEY_NONE) && ((header->func == KEYINFO_ACK) ||
+        if ((keytype != KEY_NONE) && ((header->func == CLIENT_KEY) ||
+                                      (header->func == KEYINFO_ACK) ||
                                       (header->func == FILEINFO_ACK))) {
             glog1(finfo, "Rejecting %s message from %08X: not encrypted",
                          func_name(header->func), ntohl(header->src_id));
@@ -113,12 +123,9 @@ int handle_announce_phase(unsigned char *packet, unsigned char *decrypted,
             if (*func == REGISTER) {
                 handle_open_register(message, meslen, finfo, receiver,
                                      header->src_id, regconf);
-            } else if (*func == CLIENT_KEY) {
-                handle_open_clientkey(message, meslen, finfo, receiver,
-                                      header->src_id);
             } else {
                 glog1(finfo, "Invalid function: expected "
-                             "REGISTER or CLIENT_KEY, got %s",func_name(*func));
+                             "REGISTER, got %s",func_name(*func));
             }
         } else {
             glog1(finfo, "Host %08X not in host list", ntohl(header->src_id));
@@ -422,10 +429,9 @@ void handle_transfer_phase(unsigned char *packet,unsigned char *decrypted,
             send_abort(finfo, "Not in host list", receiver, header->src_id,0,0);
             return;
         }
-        if (!validate_and_decrypt(packet, packetlen, &decrypted, &decryptlen,
-                keytype, groupkey, groupsalt, ivlen, hashtype, grouphmackey,
-                hmaclen, sigtype, keyextype, destlist[hostidx].encinfo->pubkey,
-                destlist[hostidx].encinfo->pubkeylen)) {
+        if (!validate_and_decrypt(packet, packetlen, &decrypted, 
+                &decryptlen, keytype, destlist[hostidx].encinfo->c_app_key, 
+                destlist[hostidx].encinfo->c_app_iv, ivlen)) {
             glog1(finfo, "Rejecting message from %s: decrypt/validate failed",
                          destlist[hostidx].name);
             return;
@@ -759,7 +765,7 @@ THREAD_FUNC transfer_send_thread(void *infop)
         if (block < finfo->blocks) {
             finfo->naklist[block] = 0;
         }
-        while (!finfo->naklist[block] && (block < finfo->blocks)) {
+        while ((block < finfo->blocks) && !finfo->naklist[block]) {
             block++;
         }
         if (block < finfo->blocks) {
@@ -1259,10 +1265,9 @@ void handle_completion_phase(unsigned char *packet,
             send_abort(finfo, "Not in host list", receiver, header->src_id,0,0);
             return;
         }
-        if (!validate_and_decrypt(packet, packetlen, &decrypted, &decryptlen,
-                keytype, groupkey, groupsalt, ivlen, hashtype, grouphmackey,
-                hmaclen, sigtype, keyextype, destlist[hostidx].encinfo->pubkey,
-                destlist[hostidx].encinfo->pubkeylen)) {
+        if (!validate_and_decrypt(packet, packetlen, &decrypted, 
+                &decryptlen, keytype, destlist[hostidx].encinfo->c_app_key, 
+                destlist[hostidx].encinfo->c_app_iv, ivlen)) {
             glog1(finfo, "Rejecting message from %s: "
                          "decrypt/validate failed", destlist[hostidx].name);
             return;
@@ -1271,7 +1276,7 @@ void handle_completion_phase(unsigned char *packet,
         message = decrypted;
         meslen = decryptlen;
     } else {
-        if ((keytype != KEY_NONE) && ( (header->func == STATUS) ||
+        if ((keytype != KEY_NONE) && ((header->func == STATUS) ||
                 (header->func == COMPLETE) || (header->func == ABORT))) {
             glog1(finfo, "Rejecting %s message from %08X: not encrypted",
                          func_name(header->func), ntohl(header->src_id));
@@ -1379,7 +1384,7 @@ void completion_phase(struct finfo_t *finfo)
         } 
     }
     for (i = 0; i < destcount; i++) {
-        if (destlist[i].status == DEST_ACTIVE) {
+        if (!destlist[i].isproxy && destlist[i].status == DEST_ACTIVE) {
             glog1(finfo, "Couldn't get COMPLETE for group from %s",
                          destlist[i].name);
             destlist[i].status = DEST_LOST;
