@@ -1,7 +1,7 @@
 /*
  *  UFTP - UDP based FTP with multicast
  *
- *  Copyright (C) 2001-2019   Dennis A. Bush, Jr.   bush@tcnj.edu
+ *  Copyright (C) 2001-2020   Dennis A. Bush, Jr.   bush@tcnj.edu
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -480,7 +480,7 @@ int send_files(void)
     log2(group_info.group_id, 0, 0, "%s", VERSIONSTR);
     if (!showtime) clog2(group_info.group_id, 0, 0, "Starting at %s",ctime(&t));
     if (privkey.key) {
-        if ((keyextype == KEYEX_RSA) || (keyextype == KEYEX_ECDH_RSA)) {
+        if (privkeytype == KEYBLOB_RSA) {
             log2(group_info.group_id, 0, 0,
                     "Loaded %d bit RSA key with fingerprint %s",
                        RSA_keylen(privkey.rsa) * 8,
@@ -496,7 +496,7 @@ int send_files(void)
         log2(group_info.group_id, 0, 0, "Loaded ECDH key with curve %s",
                    curve_name(get_EC_curve(dhkey.ec)));
     }
-    if (cc_type == CC_NONE || cc_type == CC_UFTP3) {
+    if (cc_type == CC_NONE) {
         if (rate == -1) {
             log2(group_info.group_id, 0, 0,
                     "Transfer rate: full interface speed");
