@@ -1,7 +1,7 @@
 /*
  *  UFTP - UDP based FTP with multicast
  *
- *  Copyright (C) 2001-2019   Dennis A. Bush, Jr.   bush@tcnj.edu
+ *  Copyright (C) 2001-2020   Dennis A. Bush, Jr.   bush@tcnj.edu
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -180,6 +180,7 @@ void send_cong_ctrl(const struct finfo_t *finfo, double l_grtt,
     struct uftp_h *header;
     struct cong_ctrl_h *cong_ctrl;
     struct timeval now;
+    int64_t now_us;
     int payloadlen, enclen;
 
     buf = safe_calloc(MAXMTU, 1); 
@@ -194,8 +195,9 @@ void send_cong_ctrl(const struct finfo_t *finfo, double l_grtt,
     cong_ctrl->cc_seq = htons(l_cc_seq);
     cong_ctrl->cc_rate = htons(quantize_rate(l_cc_rate));
     gettimeofday(&now, NULL);
-    cong_ctrl->tstamp_sec = htonl(now.tv_sec);
-    cong_ctrl->tstamp_usec = htonl(now.tv_usec);
+    now_us = tv_to_usec(now);
+    cong_ctrl->tstamp_hi = htonl((now_us & 0xFFFFFFFF00000000ULL) >> 32);
+    cong_ctrl->tstamp_lo = htonl(now_us & 0x00000000FFFFFFFFULL);
 
     bodyptr = (unsigned char *)cong_ctrl + (cong_ctrl->hlen * 4);
     memcpy(bodyptr, body, len);
@@ -204,8 +206,7 @@ void send_cong_ctrl(const struct finfo_t *finfo, double l_grtt,
     if (keytype != KEY_NONE) {
         encrypted = NULL;
         if (!encrypt_and_sign(buf, &encrypted, payloadlen, &enclen, keytype,
-                groupkey, groupsalt, &ivctr, ivlen, hashtype, grouphmackey,
-                hmaclen, sigtype, keyextype, privkey, privkeylen)) {
+                server_app_key, server_app_iv, &ivctr, ivlen)) {
             glog0(finfo, "Error encrypting CONG_CTRL");
             free(buf);
             return;
@@ -352,15 +353,18 @@ void handle_complete(const unsigned char *message, unsigned meslen,
 void handle_tfmcc_ack_info(const struct finfo_t *finfo,
                            const struct tfmcc_ack_info_he *tfmcc, int hostidx)
 {
-    struct timeval now, msgtime;
+    struct timeval now;
     int flag_ss, flag_rtt;
     int64_t client_rate, rate_1grtt;
+    int64_t msgtime_us, now_us;
     double l_adv_grtt;
 
     gettimeofday(&now, NULL);
-    msgtime.tv_sec = ntohl(tfmcc->tstamp_sec);
-    msgtime.tv_usec = ntohl(tfmcc->tstamp_usec);
-    destlist[hostidx].rtt = diff_usec(now, msgtime) / 1000000.0;
+    now_us = tv_to_usec(now);
+    msgtime_us = (int64_t)ntohl(tfmcc->tstamp_hi) << 32;
+    msgtime_us |= ntohl(tfmcc->tstamp_lo);
+
+    destlist[hostidx].rtt = (now_us - msgtime_us) / 1000000.0;
     if (destlist[hostidx].rtt < CLIENT_RTT_MIN) {
         destlist[hostidx].rtt = CLIENT_RTT_MIN;
     }
@@ -651,8 +655,7 @@ int send_data(const struct finfo_t *finfo, unsigned char *packet, int datalen,
     payloadlen = (fileseg->hlen * 4) + datalen;
     if (keytype != KEY_NONE) {
         if (!encrypt_and_sign(packet, &encpacket, payloadlen, &enclen, keytype,
-                groupkey, groupsalt, &ivctr, ivlen, hashtype, grouphmackey,
-                hmaclen, sigtype, keyextype, privkey, privkeylen)) {
+                server_app_key, server_app_iv, &ivctr, ivlen)) {
             glog0(finfo, "Error encrypting FILESEG");
             return 0;
         }
@@ -693,10 +696,10 @@ void print_status_file(const struct finfo_t *finfo, struct timeval start_time)
             } else {
                 throughput = 0;
             }
-            fprintf(status_file, "STATS;%s;%d;%d;%d;%sMB;%.3f;%.2fKB/s\n",
+            fprintf(status_file,"STATS;%s;%d;%d;%d;" F_u64 "MB;%.3f;%.2fKB/s\n",
                     destlist[i].name, destlist[i].num_copy,
                     destlist[i].num_overwrite, destlist[i].num_skip,
-                    printll(destlist[i].total_size / 1048576),
+                    destlist[i].total_size / 1048576,
                     destlist[i].total_time,
                     throughput);
         }
@@ -707,8 +710,8 @@ void print_status_file(const struct finfo_t *finfo, struct timeval start_time)
         if (destlist[i].isproxy) {
             continue;
         }
-        fprintf(status_file, "RESULT;%s;%s;%sKB;", destlist[i].name,
-                finfo->destfname, printll(finfo->size / 1024));
+        fprintf(status_file, "RESULT;%s;%s;" F_u64 "KB;", destlist[i].name,
+                finfo->destfname, finfo->size / 1024);
         switch (destlist[i].status) {
         case DEST_MUTE:
             fprintf(status_file, "mute;\n");
@@ -757,8 +760,9 @@ void print_status_file(const struct finfo_t *finfo, struct timeval start_time)
                 break;
             }
             if (destlist[i].freespace != -1) {
-                fprintf(status_file, "FREESPACE;%s;%s\n", finfo->destfname,
-                        printll(destlist[i].freespace));
+                fprintf(status_file, 
+                        "FREESPACE;%s;" F_i64 "\n", finfo->destfname,
+                        destlist[i].freespace);
             }
             break;
         default:
@@ -844,8 +848,8 @@ void print_status(const struct finfo_t *finfo, struct timeval start_time)
                 break;
             }
             if (destlist[i].freespace != -1) {
-                glog2(finfo, "  Free space on host: %s bytes",
-                             printll(destlist[i].freespace));
+                glog2(finfo, "  Free space on host: " F_i64 " bytes",
+                             destlist[i].freespace);
             }
             break;
         default:
