@@ -53,18 +53,19 @@
 #define usleep(t) Sleep((int)(t)/1000)
 #define sleep(t) Sleep((int)(t)*1000)
 #define strdup(p) _strdup(p)
-#define utime(f, t) _utime(f, t)
+#define utime(f, t) _utime64(f, t)
 #define isfullpath(str) (((str[1] == ':') && (str[2] == '\\')) || \
                          ((str[0] == '\\') && (str[1] == '\\')))
 #define PATH_SEP '\\'
 
 #define F_i64 "%I64d"
+#define F_u64 "%I64u"
 
 typedef uint64_t f_offset_t;
-typedef struct _stat32i64 stat_struct;
-typedef struct _utimbuf utim_buf;
-#define stat_func(name, buf) _stat32i64(name, buf)
-#define lstat_func(name, buf) _stat32i64(name, buf)
+typedef struct _stat64 stat_struct;
+typedef struct __utimbuf64 utim_buf;
+#define stat_func(name, buf) _stat64(name, buf)
+#define lstat_func(name, buf) _stat64(name, buf)
 #define S_ISCHR(mode)   (((mode) & S_IFMT) == S_IFCHR)
 #define S_ISDIR(mode)   (((mode) & S_IFMT) == S_IFDIR)
 #define S_ISREG(mode)   (((mode) & S_IFMT) == S_IFREG)
@@ -92,6 +93,7 @@ typedef uintptr_t thread_t;
 
 #else  // if WINDOWS
 
+#include <inttypes.h>
 #include <pthread.h>
 
 
@@ -110,7 +112,8 @@ typedef unsigned int socklen_t;
 #define PATH_SEP '/'
 #endif
 
-#define F_i64 "%lld"
+#define F_i64 "%" PRId64
+#define F_u64 "%" PRIu64
 
 typedef uint64_t f_offset_t;
 typedef struct stat stat_struct;
@@ -145,8 +148,9 @@ typedef pthread_t thread_t;
 
 #endif // if WINDOWS
 
-#define VERSIONSTR "UFTP version 4.10.2  Copyright (C) 2001-2020  Dennis A. Bush"
-#define UFTP_VER_NUM 0x40
+#define VERSIONSTR "UFTP version 5.0  Copyright (C) 2001-2020  Dennis A. Bush"
+#define UFTP_VER_NUM 0x50
+#define UFTP4_VER_NUM 0x40
 
 #define ANNOUNCE      1
 #define REGISTER      2
@@ -163,6 +167,7 @@ typedef pthread_t thread_t;
 #define DONE_CONF    13
 #define HB_REQ       14
 #define HB_RESP      15
+// obsolete, keeping for v4 compatibility
 #define KEY_REQ      16
 #define PROXY_KEY    17
 #define ENCRYPTED    18
@@ -177,8 +182,7 @@ typedef pthread_t thread_t;
 #define FTYPE_FREESPACE 4
 
 #define KEY_NONE        0
-#define KEY_DES         1
-#define KEY_DES_EDE3    2
+// reserved, obsolete: 1-4, keeping 3-4 for v4 compatibility
 #define KEY_AES128_CBC  3
 #define KEY_AES256_CBC  4
 #define KEY_AES128_GCM  5
@@ -187,17 +191,19 @@ typedef pthread_t thread_t;
 #define KEY_AES256_CCM  8
 
 #define HASH_NONE   0
-#define HASH_MD5    1
+// reserved, obsolete: 1-2, keeping 2 for v4 ECDH compatibility
 #define HASH_SHA1   2
 #define HASH_SHA256 3
 #define HASH_SHA384 4
 #define HASH_SHA512 5
 
+// obsolete, keeping for v4 compatibility
 #define SIG_NONE    0
 #define SIG_HMAC    1
 #define SIG_KEYEX   2
 #define SIG_AUTHENC 3
 
+// obsolete, keeping for v4 compatibility
 #define KEYEX_NONE          0
 #define KEYEX_RSA           1
 #define KEYEX_ECDH_RSA      2
@@ -206,36 +212,14 @@ typedef pthread_t thread_t;
 #define KEYBLOB_RSA     1
 #define KEYBLOB_EC      2
 
-#define CURVE_sect163k1     1
-#define CURVE_sect163r1     2
-#define CURVE_sect163r2     3
-#define CURVE_sect193r1     4
-#define CURVE_sect193r2     5
-#define CURVE_sect233k1     6
-#define CURVE_sect233r1     7
-#define CURVE_sect239k1     8
-#define CURVE_sect283k1     9
-#define CURVE_sect283r1     10
-#define CURVE_sect409k1     11
-#define CURVE_sect409r1     12
-#define CURVE_sect571k1     13
-#define CURVE_sect571r1     14
-#define CURVE_secp160k1     15
-#define CURVE_secp160r1     16
-#define CURVE_secp160r2     17
-#define CURVE_secp192k1     18
-#define CURVE_secp192r1     19
-#define CURVE_secp224k1     20
-#define CURVE_secp224r1     21
-#define CURVE_secp256k1     22
+// reserved, obsolete: 1-22
 #define CURVE_secp256r1     23
 #define CURVE_secp384r1     24
 #define CURVE_secp521r1     25
-#define CURVE_prime192v1    CURVE_secp192r1
 #define CURVE_prime256v1    CURVE_secp256r1
 
 #define CC_NONE     0
-#define CC_UFTP3    1
+// reserved, obsolete: 1
 #define CC_TFMCC    2
 #define CC_PGMCC    3
 
@@ -287,15 +271,20 @@ typedef pthread_t thread_t;
 #define MAX_INTERFACES 200
 #define MAXMTU 9000
 
-#define PUBKEY_LEN 264  // big enough for a keyblob with RSA-2048
-#define RAND_LEN 32     // RFC 5246
+#define PUBKEY_LEN 392  // big enough for a keyblob with RSA-3072
+#define RAND_LEN 32     // RFC 8446/5246
 #define HMAC_LEN 64     // big enough for SHA-512
-#define VERIFY_LEN 12   // RFC 5246
-#define MASTER_LEN 48   // RFC 5246
+#define HASH_LEN 64     // big enough for SHA-512
+#define VERIFY4_LEN 12  // RFC 5246
+#define MASTER4_LEN 48  // RFC 5246
+#define MASTER_LEN 32   // based on SHA-256 length
+#define ENC_MASTER_LEN 48 // based on SHA-256 length + tag length
 #define MAXIV 16        // big enough for AES256
 #define MAXKEY 32       // big enough for AES256
 #define KEYBLSIZE 16    // Maximum symmetric key blocksize
-#define DEF_RSA_LEN 512 // Default length of generated RSA keys
+#define DEF_RSA_LEN 1024 // Default length of generated RSA keys
+#define RSA_MIN 1024    // Minimum RSA key length
+#define RSA_MAX 3072    // Maximum RSA key length
 #define DEF_CURVE CURVE_prime256v1  // Default EC curve
 #define RSA_EXP 65537   // Public key exponent of generated RSA keys
 #define SALT_LEN 4      // Length of salt for IV
@@ -331,6 +320,7 @@ struct uftp_h {
 struct encrypted_h {
     uint32_t iv_ctr_hi;
     uint32_t iv_ctr_lo;
+    // v4 only, set to 0 in v5
     uint16_t sig_len;
     uint16_t payload_len;
 };  // sizeof = 12 + sig_len + payload_len
@@ -343,8 +333,8 @@ struct announce_h {
     uint8_t cc_type;
     uint8_t reserved;
     uint16_t blocksize;
-    uint32_t tstamp_sec;
-    uint32_t tstamp_usec;
+    uint32_t tstamp_hi;
+    uint32_t tstamp_lo;
     //uint32_t publicmcast;   // for IPv4
     //uint8_t publicmcast[16];  // for IPv6
     //uint32_t privatemcast;   // for IPv4
@@ -355,6 +345,7 @@ struct enc_info_he {
     uint8_t exttype;   // always EXT_ENC_INFO
     uint8_t extlen;
     uint8_t flags;
+    // v4 only, set to 0 in v5
     uint8_t keyextype_sigtype;  // & 0xF0 = keyextype, & 0x0F = sigtype
     uint8_t keytype;
     uint8_t hashtype;
@@ -395,12 +386,12 @@ struct client_key_h {
 struct register_h {
     uint8_t func;  // always REGISTER
     uint8_t hlen;
-    uint16_t keyinfo_len;
-    uint32_t tstamp_sec;
-    uint32_t tstamp_usec;
+    uint16_t dhlen;
+    uint32_t tstamp_hi;
+    uint32_t tstamp_lo;
     uint8_t rand2[RAND_LEN];
-    //uint8_t keyinfo[];  // Either an EC key or an RSA encrypted premaster
-};  // sizeof = 44 + (server_key_len or ec_len) + {uint32_t[]}
+    //uint8_t dhkey[];
+};  // sizeof = 44 + dhlen + {uint32_t[]}
 
 struct regconf_h {
     uint8_t func;  // always REG_CONF
@@ -411,22 +402,22 @@ struct regconf_h {
 struct keyinfo_h {
     uint8_t func;  // always KEYINFO
     uint8_t hlen;
-    uint16_t reserved;
+    uint16_t siglen;
     uint32_t iv_ctr_hi;
     uint32_t iv_ctr_lo;
-};  // sizeof = 12 + {destkey[]}
+};  // sizeof = 12 + siglen + {destkey[]}
 
 struct destkey {
     uint32_t dest_id;
-    uint8_t groupmaster[MASTER_LEN];      // based on 16 byte blocksize
-};  // sizeof = 52 */
+    uint8_t groupmaster[ENC_MASTER_LEN];
+};  // sizeof = 52
 
 struct keyinfoack_h {
     uint8_t func;  // always KEYINFO_ACK
     uint8_t hlen;
-    uint16_t reserved;
-    uint8_t verify_data[VERIFY_LEN];
-};  // sizeof = 16
+    uint16_t verify_len;
+    //uint8_t verify_data[];
+};  // sizeof = 4 + verify_len
 
 struct fileinfo_h {
     uint8_t func;  // always FILEINFO
@@ -434,14 +425,14 @@ struct fileinfo_h {
     uint16_t file_id;
     uint8_t ftype;
     uint8_t reserved1;
-    uint16_t reserved2;
+    uint16_t ftstamp_hi;
     uint8_t namelen;
     uint8_t linklen;
     uint16_t hifsize;
     uint32_t lofsize;
-    uint32_t ftstamp;
-    uint32_t tstamp_sec;
-    uint32_t tstamp_usec;
+    uint32_t ftstamp_lo;
+    uint32_t tstamp_hi;
+    uint32_t tstamp_lo;
     //char name[MAXPATHNAME];
 };  // sizeof = 28 + namelen + {uint32_t[]}
 
@@ -452,8 +443,8 @@ struct fileinfoack_h {
     uint8_t flags;
     uint8_t reserved1;
     uint16_t reserved2;
-    uint32_t tstamp_sec;
-    uint32_t tstamp_usec;
+    uint32_t tstamp_hi;
+    uint32_t tstamp_lo;
 };  // sizeof = 16 + {uint32_t[]}
 
 struct fileseg_h {
@@ -506,8 +497,8 @@ struct tfmcc_ack_info_he {
     uint16_t cc_seq;
     uint16_t cc_rate;
     uint32_t client_id;
-    uint32_t tstamp_sec;
-    uint32_t tstamp_usec;
+    uint32_t tstamp_hi;
+    uint32_t tstamp_lo;
 };  // sizeof = 20 
 
 // Append to STATUS
@@ -515,8 +506,8 @@ struct pgmcc_nak_info_he {
     uint8_t exttype;   // always EXT_PGMCC_NAK_INFO
     uint8_t extlen;
     uint16_t loss;
-    uint32_t tstamp_sec;
-    uint32_t tstamp_usec;
+    uint32_t tstamp_hi;
+    uint32_t tstamp_lo;
 };  // sizeof = 12
 
 // Append to CC_ACK
@@ -527,8 +518,8 @@ struct pgmcc_ack_info_he {
     uint16_t loss;
     uint16_t max_seq;
     uint32_t ack_bitmap;
-    uint32_t tstamp_sec;
-    uint32_t tstamp_usec;
+    uint32_t tstamp_hi;
+    uint32_t tstamp_lo;
 };  // sizeof = 20
 
 struct complete_h {
@@ -560,7 +551,7 @@ struct abort_h {
     uint8_t flags;
     uint8_t reserved;
     uint32_t host;
-    char message[300]; // TODO: define error codes
+    char message[300];
 };  // sizeof = 308
 
 struct hb_req_h {
@@ -588,7 +579,7 @@ struct key_req_h {
     uint16_t reserved;
 };  // sizeof = 4
 
-struct proxy_key_h {
+struct proxy_key_4_h {
     uint8_t func;  // always PROXY_KEY
     uint8_t hlen;
     uint16_t bloblen;
@@ -600,14 +591,25 @@ struct proxy_key_h {
     //uint8_t verify[];
 };  // sizeof = 12 + bloblen + dhlen + siglen
 
+struct proxy_key_h {
+    uint8_t func;  // always PROXY_KEY
+    uint8_t hlen;
+    uint16_t bloblen;
+    uint16_t dhlen;
+    uint16_t siglen;
+    //uint8_t keyblob[];
+    //uint8_t dhkey[];
+    //uint8_t signature[];
+};  // sizeof = 8 + bloblen + dhlen + siglen
+
 struct cong_ctrl_h {
     uint8_t func;  // always CONG_CTRL
     uint8_t hlen;
     uint16_t reserved;
     uint16_t cc_seq;
     uint16_t cc_rate;
-    uint32_t tstamp_sec;
-    uint32_t tstamp_usec;
+    uint32_t tstamp_hi;
+    uint32_t tstamp_lo;
 }; // sizeof = 16 + {cc_item[]}
 
 struct cc_item {
