@@ -222,19 +222,19 @@ pid_t GENERIC_SETSID(void) { return(0); }
 
 #endif /*if WINDOWS*/
 
-const char *printll(int64_t n)
+struct timeval usec_to_tv(int64_t t)
 {
-    static char str[50];
-
-#ifdef WINDOWS
-    snprintf(str, sizeof(str) - 1, "%I64d", n);
-#else
-    snprintf(str, sizeof(str) - 1, "%lld", (long long)n);
-#endif
-    str[sizeof(str) - 1] = '\x0';
-    return &str[0];
+    struct timeval tv;
+    // TODO: Y2038 issue, switch to timespec / clock_gettime
+    tv.tv_sec = (long)(t / 1000000);
+    tv.tv_usec = t % 1000000;
+    return tv;
 }
 
+int64_t tv_to_usec(struct timeval tv)
+{
+    return (int64_t)tv.tv_sec * 1000000 + tv.tv_usec;
+}
 
 int32_t diff_sec(struct timeval t2, struct timeval t1)
 {
@@ -357,50 +357,6 @@ const char *func_name(int func)
 const char *curve_name(int curve)
 {
     switch (curve) {
-    case CURVE_sect163k1:
-        return "sect163k1";
-    case CURVE_sect163r1:
-        return "sect163r1";
-    case CURVE_sect163r2:
-        return "sect163r2";
-    case CURVE_sect193r1:
-        return "sect193r1";
-    case CURVE_sect193r2:
-        return "sect193r2";
-    case CURVE_sect233k1:
-        return "sect233k1";
-    case CURVE_sect233r1:
-        return "sect233r1";
-    case CURVE_sect239k1:
-        return "sect239k1";
-    case CURVE_sect283k1:
-        return "sect283k1";
-    case CURVE_sect283r1:
-        return "sect283r1";
-    case CURVE_sect409k1:
-        return "sect409k1";
-    case CURVE_sect409r1:
-        return "sect409r1";
-    case CURVE_sect571k1:
-        return "sect571k1";
-    case CURVE_sect571r1:
-        return "sect571r1";
-    case CURVE_secp160k1:
-        return "secp160k1";
-    case CURVE_secp160r1:
-        return "secp160r1";
-    case CURVE_secp160r2:
-        return "secp160r2";
-    case CURVE_secp192k1:
-        return "secp192k1";
-    case CURVE_secp192r1:
-        return "prime192v1";
-    case CURVE_secp224k1:
-        return "secp224k1";
-    case CURVE_secp224r1:
-        return "secp224r1";
-    case CURVE_secp256k1:
-        return "secp256k1";
     case CURVE_secp256r1:
         return "prime256v1";
     case CURVE_secp384r1:
@@ -418,56 +374,12 @@ const char *curve_name(int curve)
  */
 uint8_t get_curve(const char *name)
 {
-    if (!strcmp(name, "sect163k1")) {
-        return CURVE_sect163k1;
-    } else if (!strcmp(name, "sect163r1")) {
-        return CURVE_sect163r1;
-    } else if (!strcmp(name, "sect163r2")) {
-        return CURVE_sect163r2;
-    } else if (!strcmp(name, "sect193r1")) {
-        return CURVE_sect193r1;
-    } else if (!strcmp(name, "sect193r2")) {
-        return CURVE_sect193r2;
-    } else if (!strcmp(name, "sect233k1")) {
-        return CURVE_sect233k1;
-    } else if (!strcmp(name, "sect233r1")) {
-        return CURVE_sect233r1;
-    } else if (!strcmp(name, "sect239k1")) {
-        return CURVE_sect239k1;
-    } else if (!strcmp(name, "sect283k1")) {
-        return CURVE_sect283k1;
-    } else if (!strcmp(name, "sect283r1")) {
-        return CURVE_sect283r1;
-    } else if (!strcmp(name, "sect409k1")) {
-        return CURVE_sect409k1;
-    } else if (!strcmp(name, "sect409r1")) {
-        return CURVE_sect409r1;
-    } else if (!strcmp(name, "sect571k1")) {
-        return CURVE_sect571k1;
-    } else if (!strcmp(name, "sect571r1")) {
-        return CURVE_sect571r1;
-    } else if (!strcmp(name, "secp160k1")) {
-        return CURVE_secp160k1;
-    } else if (!strcmp(name, "secp160r1")) {
-        return CURVE_secp160r1;
-    } else if (!strcmp(name, "secp160r2")) {
-        return CURVE_secp160r2;
-    } else if (!strcmp(name, "secp192k1")) {
-        return CURVE_secp192k1;
-    } else if (!strcmp(name, "secp192r1")) {
-        return CURVE_secp192r1;
-    } else if (!strcmp(name, "secp224r1")) {
-        return CURVE_secp224r1;
-    } else if (!strcmp(name, "secp256k1")) {
-        return CURVE_secp256k1;
-    } else if (!strcmp(name, "secp256r1")) {
+    if (!strcmp(name, "secp256r1")) {
         return CURVE_secp256r1;
     } else if (!strcmp(name, "secp384r1")) {
         return CURVE_secp384r1;
     } else if (!strcmp(name, "secp521r1")) {
         return CURVE_secp521r1;
-    } else if (!strcmp(name, "prime192v1")) {
-        return CURVE_prime192v1;
     } else if (!strcmp(name, "prime256v1")) {
         return CURVE_prime256v1;
     } else {
@@ -762,7 +674,7 @@ void split_path(const char *path, char **dir, char **file)
  */
 int parse_fingerprint(unsigned char *fingerprint, const char *fingerprint_str)
 {
-    char *p, *tmp;
+    char *p, *tmp, *saveptr;
     int num, len;
 
     if (fingerprint_str == NULL) {
@@ -770,7 +682,8 @@ int parse_fingerprint(unsigned char *fingerprint, const char *fingerprint_str)
     }
     tmp = strdup(fingerprint_str);
     len = 0;
-    p = strtok(tmp, ":");
+    saveptr = NULL;
+    p = strtok_r(tmp, ":", &saveptr);
     if (p == NULL) {
         log1(0, 0, 0, "Invalid fingerprint %s", fingerprint_str);
         free(tmp);
@@ -796,12 +709,28 @@ int parse_fingerprint(unsigned char *fingerprint, const char *fingerprint_str)
             return 0;
         }
         fingerprint[len++] = (uint8_t)num;
-        p = strtok(NULL, ":");
+        p = strtok_r(NULL, ":", &saveptr);
     } while (p);
     free(tmp);
     return 1;
 }
 
+/**
+ * Looks up a host in a list of fingerprint structs
+ * Returns NULL if not found
+ */
+struct fp_list_t *fp_lookup(uint32_t id, struct fp_list_t* list, int count)
+{
+    int i;
+
+    for (i = 0; i < count; i++) {
+        if (list[i].uid == id) {
+            return &list[i];
+        }
+    }
+    return NULL;
+}
+
 /**
  * Tests a sockaddr_u union to see if it's a valid multicast address
  */
@@ -887,7 +816,7 @@ uint64_t uftp_htonll(uint64_t val)
 
     p = (unsigned char *)&rval;
     for (i = 0; i < 8; i++) {
-        p[7 - i] = (val & (0xFFLL << (i * 8))) >> (i * 8);
+        p[7 - i] = (val & (0xFFULL << (i * 8))) >> (i * 8);
     }
     return rval;
 }
@@ -947,6 +876,18 @@ int conn_reset_err(void)
 #endif
 }
 
+/**
+ * Returns whether a call was interrupted
+ */
+int interrupted_err(void)
+{
+#ifdef WINDOWS
+    return (WSAGetLastError() == WSAEINTR);
+#else
+    return (errno == EINTR);
+#endif
+}
+
 /**
  * Calls sendto, retrying if the send would block.
  * The calling function should check for and log any other errors.
@@ -1112,7 +1053,9 @@ int read_packet(SOCKET sock, union sockaddr_u *su, unsigned char *buffer,
         if (timeout) tv = *timeout;
         if ((rval = select(FD_SETSIZE-1, &fdin, NULL, NULL,
                            (timeout ? &tv : NULL))) == SOCKET_ERROR) {
-            sockerror(0, 0, 0, "Select failed");
+            if (!interrupted_err()) {
+                sockerror(0, 0, 0, "Select failed");
+            }
             return -1;
         }
         if (rval == 0) {
@@ -1143,8 +1086,8 @@ void memxor(void *p1, const void *p2, int len)
  * For a 64-bit IV (DES, 3DES):     IV = (S + src_ID) XOR ctr
  * All values other should be in network byte order.
  */
-void build_iv(uint8_t *iv, const uint8_t *salt, int ivlen, uint64_t ivctr,
-              uint32_t src_id)
+void build_iv4(uint8_t *iv, const uint8_t *salt, int ivlen, uint64_t ivctr,
+               uint32_t src_id)
 {
     char tmp[16], tmp2[16];
     int tmplen, tmp2len;
@@ -1178,6 +1121,16 @@ void build_iv(uint8_t *iv, const uint8_t *salt, int ivlen, uint64_t ivctr,
     memcpy(iv, tmp, tmplen);
 }
 
+/**
+ * Builds an IV as: S XOR CTR
+ * All values should be in network byte order.
+ */
+void build_iv(uint8_t *iv, const uint8_t *salt, int ivlen, uint64_t ivctr)
+{
+    memcpy(iv, salt, ivlen);
+    memxor(iv + (ivlen - sizeof(uint64_t)), &ivctr, sizeof(uint64_t));
+}
+
 /**
  * Outputs data buffers to log in hex.
  * Used only for debugging
@@ -1245,85 +1198,29 @@ int unauth_key(int keytype)
 int validate_and_decrypt(unsigned char *encpacket, unsigned int enclen,
                          unsigned char **decpacket, unsigned int *declen,
                          int keytype, const uint8_t *key,
-                         const uint8_t *salt, int ivlen, int hashtype,
-                         const uint8_t *hmackey, uint8_t hmaclen, int sigtype,
-                         int keyextype, union key_t pubkey, int pubkeylen)
+                         const uint8_t *salt, int ivlen)
 {
     struct uftp_h *header;
     struct encrypted_h *encrypted;
-    unsigned char *payload, *sig, *sigcopy, *sigtest, *iv;
-    unsigned int hsiglen, siglen, len, rval, allocdec;
+    unsigned char *payload, *iv;
+    unsigned int rval, allocdec;
     uint64_t ivctr;
 
-    if (sigtype == SIG_HMAC) {
-        siglen = hmaclen;
-    } else if (sigtype == SIG_KEYEX) {
-        siglen = pubkeylen;
-    } else if (sigtype == SIG_AUTHENC) {
-        siglen = 0;
-    } else {
-        log0(0, 0, 0, "Invalid signature type");
-        return 0;
-    }
-
     header = (struct uftp_h *)encpacket;
     encrypted = (struct encrypted_h *)(encpacket + sizeof(struct uftp_h));
-    hsiglen = ntohs(encrypted->sig_len);
-    sig = (unsigned char *)encrypted + sizeof(struct encrypted_h);
-    payload = sig + hsiglen;
+    payload = (unsigned char *)encrypted + sizeof(struct encrypted_h);
 
     if (header->func != ENCRYPTED) {
         log0(0, 0, 0, "Attempt to decrypt non-encrypted message");
         return 0;
     }
     if (enclen != (sizeof(struct uftp_h) + sizeof(struct encrypted_h) +
-            hsiglen + ntohs(encrypted->payload_len))) {
+            ntohs(encrypted->payload_len))) {
         log0(0, 0, 0, "Invalid signature and/or encrypted payload length");
         return 0;
     }
-    if (hsiglen != siglen) {
-        log0(0, 0, 0, "Signature length incorrect: got %d, expected %d",
-                      hsiglen, siglen);
-        return 0;
-    }
 
-    sigcopy = safe_calloc(siglen, 1);
-    sigtest = safe_calloc(siglen, 1);
     iv = safe_calloc(ivlen, 1);
-
-    if (sigtype != SIG_AUTHENC) {
-        memcpy(sigcopy, sig, hsiglen);
-        memset(sig, 0, hsiglen);
-    }
-    if (sigtype == SIG_HMAC) {
-        if (!create_hmac(hashtype, hmackey, hsiglen, encpacket, enclen,
-                         sigtest, &len)) {
-            log0(0, 0, 0, "HMAC creation failed");
-            rval = 0;
-            goto end;
-        }
-        if (memcmp(sigtest, sigcopy, len)) {
-            log0(0, 0, 0, "HMAC verification failed");
-            rval = 0;
-            goto end;
-        }
-    } else if ((sigtype == SIG_KEYEX) && (keyextype == KEYEX_ECDH_ECDSA)) {
-        if (!verify_ECDSA_sig(pubkey.ec, hashtype, encpacket, enclen,
-                              sigcopy, ntohs(encrypted->sig_len))) {
-            log0(0, 0, 0, "ECDSA signature verification failed");
-            rval = 0;
-            goto end;
-        }
-    } else if ((sigtype == SIG_KEYEX) &&
-            ((keyextype == KEYEX_RSA) || (keyextype == KEYEX_ECDH_RSA))) {
-        if (!verify_RSA_sig(pubkey.rsa, hashtype, encpacket, enclen,
-                            sigcopy, ntohs(encrypted->sig_len))) {
-            log0(0, 0, 0, "RSA signature verification failed");
-            rval = 0;
-            goto end;
-        }
-    }
-
     allocdec = 0;
     if (*decpacket == NULL) {
         allocdec = 1;
@@ -1331,7 +1228,11 @@ int validate_and_decrypt(unsigned char *encpacket, unsigned int enclen,
     }
     ivctr = ntohl(encrypted->iv_ctr_lo);
     ivctr |= (uint64_t)ntohl(encrypted->iv_ctr_hi) << 32;
-    build_iv(iv, salt, ivlen, uftp_htonll(ivctr), header->src_id);
+    if (header->version == UFTP4_VER_NUM) {
+        build_iv4(iv, salt, ivlen, uftp_htonll(ivctr), header->src_id);
+    } else {
+        build_iv(iv, salt, ivlen, uftp_htonll(ivctr));
+    }
     if (!decrypt_block(keytype, iv, key, encpacket,
             sizeof(struct uftp_h) + sizeof(struct encrypted_h),
             payload, ntohs(encrypted->payload_len), *decpacket, declen)) {
@@ -1347,8 +1248,6 @@ int validate_and_decrypt(unsigned char *encpacket, unsigned int enclen,
     rval = 1;
 
 end:
-    free(sigcopy);
-    free(sigtest);
     free(iv);
     return rval;
 }
@@ -1360,27 +1259,13 @@ end:
  */
 int encrypt_and_sign(const unsigned char *decpacket, unsigned char **encpacket,
                      int declen, int *enclen, int keytype, uint8_t *key, 
-                     const uint8_t *salt, uint64_t *ivctr, int ivlen,
-                     int hashtype, uint8_t *hmackey, uint8_t hmaclen,
-                     int sigtype, int keyextype, union key_t privkey,
-                     int privkeylen)
+                     const uint8_t *salt, uint64_t *ivctr, int ivlen)
 {
     struct uftp_h *header;
     struct encrypted_h *encrypted;
     const unsigned char *mheader;
-    unsigned char *payload, *sig, *sigcopy, *iv;
-    unsigned int payloadlen, siglen, len, allocenc;
-
-    if (sigtype == SIG_HMAC) {
-        siglen = hmaclen;
-    } else if (sigtype == SIG_KEYEX) {
-        siglen = privkeylen;
-    } else if (sigtype == SIG_AUTHENC) {
-        siglen = 0;
-    } else {
-        log0(0, 0, 0, "Invalid signature type");
-        return 0;
-    }
+    unsigned char *payload, *iv;
+    unsigned int payloadlen, allocenc;
 
     allocenc = 0;
     if (*encpacket == NULL) {
@@ -1392,24 +1277,27 @@ int encrypt_and_sign(const unsigned char *decpacket, unsigned char **encpacket,
     mheader = decpacket + sizeof(struct uftp_h);
     header = (struct uftp_h *)*encpacket;
     encrypted = (struct encrypted_h *)(*encpacket + sizeof(struct uftp_h));
-    sig = (unsigned char *)encrypted + sizeof(struct encrypted_h);
-    payload = sig + siglen;
+    payload = (unsigned char *)encrypted + sizeof(struct encrypted_h);
 
     (*ivctr)++;
     memcpy(*encpacket, decpacket, sizeof(struct uftp_h));
     header->func = ENCRYPTED;
-    encrypted->iv_ctr_hi = htonl((*ivctr & 0xFFFFFFFF00000000LL) >> 32);
-    encrypted->iv_ctr_lo = htonl(*ivctr & 0x00000000FFFFFFFFLL);
-    encrypted->sig_len = htons(siglen);
-    if (sigtype == SIG_AUTHENC) {
-        if (is_gcm_mode(keytype)) {
-            encrypted->payload_len = htons(declen + GCM_TAG_LEN);
-        } else if (is_ccm_mode(keytype)) {
-            encrypted->payload_len = htons(declen + CCM_TAG_LEN);
-        }
+    encrypted->iv_ctr_hi = htonl((*ivctr & 0xFFFFFFFF00000000ULL) >> 32);
+    encrypted->iv_ctr_lo = htonl(*ivctr & 0x00000000FFFFFFFFULL);
+    if (is_gcm_mode(keytype)) {
+        encrypted->payload_len = htons(declen + GCM_TAG_LEN);
+    } else if (is_ccm_mode(keytype)) {
+        encrypted->payload_len = htons(declen + CCM_TAG_LEN);
+    } else {
+        log0(0, 0, 0, "Invalid cipher mode for keytype %d", keytype);
+        return 0;
     }
 
-    build_iv(iv, salt, ivlen, uftp_htonll(*ivctr), header->src_id);
+    if (header->version == UFTP4_VER_NUM) {
+        build_iv4(iv, salt, ivlen, uftp_htonll(*ivctr), header->src_id);
+    } else {
+        build_iv(iv, salt, ivlen, uftp_htonll(*ivctr));
+    }
     if (!encrypt_block(keytype, iv, key, *encpacket,
             sizeof(struct uftp_h) + sizeof(struct encrypted_h),
             mheader, declen, payload, &payloadlen)) {
@@ -1422,67 +1310,19 @@ int encrypt_and_sign(const unsigned char *decpacket, unsigned char **encpacket,
         return 0;
     }
     free(iv);
-    if (sigtype != SIG_AUTHENC) {
-        encrypted->payload_len = htons(payloadlen);
-    } else {
-        if (payloadlen != ntohs(encrypted->payload_len)) {
-            log0(0, 0, 0, "Invalid payloadlen: got %d, expected %d",
-                          payloadlen, ntohs(encrypted->payload_len));
-            return 0;
-        }
+    if (payloadlen != ntohs(encrypted->payload_len)) {
+        log0(0, 0, 0, "Invalid payloadlen: got %d, expected %d",
+                      payloadlen, ntohs(encrypted->payload_len));
+        return 0;
     }
-    *enclen = sizeof(struct encrypted_h) + payloadlen + siglen;
+    *enclen = sizeof(struct encrypted_h) + payloadlen;
 
-    sigcopy = safe_calloc(siglen, 1);
-    if (sigtype != SIG_AUTHENC) {
-        memset(sig, 0, siglen);
-    }
-    if (sigtype == SIG_HMAC) {
-        if (!create_hmac(hashtype, hmackey, siglen, *encpacket,
-                         sizeof(struct uftp_h) + *enclen, sigcopy, &len)) {
-            log0(0, 0, 0, "HMAC creation failed");
-            free(sigcopy);
-            if (allocenc) {
-                free(*encpacket);
-                *encpacket = NULL;
-            }
-            return 0;
-        }
-    } else if ((sigtype == SIG_KEYEX) && (keyextype == KEYEX_ECDH_ECDSA)) {
-        if (!create_ECDSA_sig(privkey.ec, hashtype, *encpacket,
-                             sizeof(struct uftp_h) + *enclen, sigcopy, &len)) {
-            // Called function should log
-            free(sigcopy);
-            if (allocenc) {
-                free(*encpacket);
-                *encpacket = NULL;
-            }
-            return 0;
-        }
-    } else if ((sigtype == SIG_KEYEX) &&
-            ((keyextype == KEYEX_RSA) || (keyextype == KEYEX_ECDH_RSA))) {
-        if (!create_RSA_sig(privkey.rsa, hashtype, *encpacket,
-                            sizeof(struct uftp_h) + *enclen, sigcopy, &len)) {
-            // Called function should log
-            free(sigcopy);
-            if (allocenc) {
-                free(*encpacket);
-                *encpacket = NULL;
-            }
-            return 0;
-        }
-    }
-    if (sigtype != SIG_AUTHENC) {
-        memcpy(sig, sigcopy, len);
-    }
-
-    free(sigcopy);
     return 1;
 }
 
 /**
  * Pseudo-random function for an individual hashing algorithm
- * as defined in RFC 4346 and RFC 5246
+ * as defined in RFC 5246
  */
 static void P_hash(int hashtype, int bytes, 
                    const unsigned char *secret, int secret_len, 
@@ -1500,7 +1340,7 @@ static void P_hash(int hashtype, int bytes,
     *outbuf_len = 0;
     newseed_len = 0;
     memcpy(newseed, label, strlen(label));
-    newseed_len += strlen(label);
+    newseed_len += (unsigned)strlen(label);
     memcpy(newseed + newseed_len, seed, seed_len);
     newseed_len += seed_len;
 
@@ -1508,8 +1348,6 @@ static void P_hash(int hashtype, int bytes,
     inbuf_len = newseed_len;
     while (*outbuf_len < bytes)
     {
-        // TODO: create_hmac can fail under CrypoAPI.
-        // Figure out how to handle this.
         create_hmac(hashtype, secret, secret_len, inbuf, inbuf_len,
                     tmpbuf, &tmpbuf_len);
         memcpy(tmpbuf + tmpbuf_len, newseed, newseed_len);
@@ -1528,36 +1366,293 @@ static void P_hash(int hashtype, int bytes,
 
 /**
  * Pseudo-random function
- * as defined in RFC 4346 and RFC 5246
+ * as defined in RFC 5246
  */
 void PRF(int hashtype, int bytes, const unsigned char *secret, int secret_len, 
          const char *label, const unsigned char *seed, int seed_len,
          unsigned char *outbuf, int *outbuf_len)
 {
-    int i, s_len, sha_buf_len, md5_buf_len;
-    unsigned char *sha_buf, *md5_buf;
+    P_hash(hashtype, bytes, secret, secret_len, label,
+           seed, seed_len, outbuf, outbuf_len);
+}
 
-    if (hashtype == HASH_SHA1) {
-        // TLS 1.1 MD5/SHA1 combination
-        md5_buf = safe_calloc(bytes + get_hash_len(HASH_MD5), 1);
-        sha_buf = safe_calloc(bytes + get_hash_len(HASH_SHA1), 1);
-        // if secret_len is even integer division truncates the result
-        // if secret_len is odd, the + 1 effectively rounds up
-        s_len = (secret_len + 1) / 2;
-        P_hash(HASH_MD5, bytes, secret, s_len, label,
-               seed, seed_len, md5_buf, &md5_buf_len);
-        P_hash(HASH_SHA1, bytes, secret + (secret_len - s_len), s_len, label,
-               seed, seed_len, sha_buf, &sha_buf_len);
-        for (i = 0; i < bytes; i++) {
-            outbuf[i] = md5_buf[i] ^ sha_buf[i];
-        }
-        *outbuf_len = bytes;
-        free(md5_buf);
-        free(sha_buf);
-    } else {
-        P_hash(hashtype, bytes, secret, secret_len, label,
-               seed, seed_len, outbuf, outbuf_len);
+/**
+  * Creates Server_HS_Context 
+  * All integer values within the context should be in network byte order
+  */
+void create_server_context(uint32_t group_id, uint8_t group_inst,
+                           uint32_t server_id,const struct enc_info_he *encinfo,
+                           int extlen, uint8_t **context, int *context_len)
+{
+    *context = safe_malloc(sizeof(group_id) + sizeof(group_inst) +
+                           sizeof(server_id) + extlen);
+    *context_len = 0;
+    memcpy(*context + *context_len, &group_id, sizeof(group_id));
+    *context_len += sizeof(group_id);
+    memcpy(*context + *context_len, &group_inst, sizeof(group_inst));
+    *context_len += sizeof(group_inst);
+    memcpy(*context + *context_len, &server_id, sizeof(server_id));
+    *context_len += sizeof(server_id);
+    memcpy(*context + *context_len, encinfo, extlen);
+    *context_len += extlen;
+}
+
+/**
+  * Creates Proxy_HS_Context
+  * All integer values within the context should be in network byte order
+  */
+void create_proxy_context(uint32_t proxy_id, const struct proxy_key_h *proxykey,
+                          uint8_t **context, int *context_len)
+{
+    uint8_t *newcontext = safe_malloc(sizeof(proxy_id) + proxykey->hlen * 4);
+    *context = newcontext;
+    memcpy(*context + *context_len, &proxy_id, sizeof(proxy_id));
+    *context_len += sizeof(proxy_id);
+    memcpy(*context + *context_len, proxykey, proxykey->hlen * 4);
+    *context_len += proxykey->hlen * 4;
+}
+
+/**
+  * Creates Client_HS_Context1
+  * All integer within the context values should be in network byte order
+  * client_dh is an EC keyblob
+  */
+void create_client_context_1(const uint8_t *s_context, int s_context_len,
+                             const uint8_t *p_context, int p_context_len,
+                             uint32_t client_id, const uint8_t *client_dh,
+                             int client_dh_len, const uint8_t *client_rand, 
+                             uint8_t **context, int *context_len)
+{
+    *context = safe_malloc(s_context_len + p_context_len + sizeof(client_id) +
+                           client_dh_len + RAND_LEN);
+    *context_len = 0;
+    memcpy(*context + *context_len, s_context, s_context_len);
+    *context_len += s_context_len;
+    if (p_context_len) {
+        memcpy(*context + *context_len, p_context, p_context_len);
+        *context_len += p_context_len;
     }
+    memcpy(*context + *context_len, &client_id, sizeof(client_id));
+    *context_len += sizeof(client_id);
+    memcpy(*context + *context_len, client_dh, client_dh_len);
+    *context_len += client_dh_len;
+    memcpy(*context + *context_len, client_rand, RAND_LEN);
+    *context_len += RAND_LEN;
+}
+
+/**
+  * Creates Client_HS_Context2
+  * All integer values within the context should be in network byte order
+  */
+void create_client_context_2(const uint8_t *c_context1, int c_context1_len,
+                             const struct client_key_h *ckheader,int header_len,
+                             uint8_t **context, int *context_len)
+{
+    *context = safe_malloc(c_context1_len + header_len);
+    *context_len = 0;
+    memcpy(*context + *context_len, c_context1, c_context1_len);
+    *context_len += c_context1_len;
+    if (header_len) {
+        memcpy(*context + *context_len, ckheader, header_len);
+        *context_len += header_len;
+    }
+}
+
+/**
+ * HMAC based Key Derivation Function (HKDF) - Extract
+ * as defined in RFC 5869
+ */
+void HKDF_Extract(int hashtype, 
+                  const unsigned char *salt, unsigned int salt_len,
+                  const unsigned char *secret, unsigned int secret_len,
+                  unsigned char *outbuf, unsigned int *outbuf_len)
+{
+    create_hmac(hashtype, salt ? salt : (const unsigned char *)"", 
+                salt ? salt_len : 0, secret, secret_len, outbuf, outbuf_len);
+}
+
+/**
+ * HMAC based Key Derivation Function (HKDF) - Extract
+ * as defined in RFC 5869
+ * This can generate more bytes than requested, up to the hash length
+ */
+void HKDF_Expand(int hashtype, unsigned int bytes,
+                 const unsigned char *secret, unsigned int secret_len,
+                 const unsigned char *info, unsigned int info_len,
+                 unsigned char *outbuf, unsigned int *outbuf_len)
+{
+    unsigned char *inbuf, *tmpbuf;
+    unsigned inbuf_len, tmpbuf_len;
+    unsigned char counter = 1;
+
+    inbuf = safe_malloc(get_hash_len(hashtype) + info_len + 1);
+    tmpbuf = safe_malloc(get_hash_len(hashtype) + info_len + 1);
+
+    *outbuf_len = 0;
+    tmpbuf_len = 0;
+    while (*outbuf_len < bytes) {
+        memcpy(inbuf, tmpbuf, tmpbuf_len);
+        inbuf_len = tmpbuf_len;
+        memcpy(inbuf + inbuf_len, info, info_len);
+        inbuf_len += info_len;
+        memcpy(inbuf + inbuf_len, &counter, 1);
+        inbuf_len += 1;
+        create_hmac(hashtype, secret, secret_len, inbuf, inbuf_len,
+                    tmpbuf, &tmpbuf_len);
+        memcpy(outbuf + *outbuf_len, tmpbuf, tmpbuf_len);
+        *outbuf_len += tmpbuf_len;
+        counter++;
+    }
+    free(inbuf);
+    free(tmpbuf);
+}
+
+/**
+ * HDKF Expand with label
+ * derived from RFC 8446
+ * This can generate more bytes than requested, up to the hash length
+ */
+void HKDF_Expand_Label(int hashtype, unsigned int bytes, const char *label,
+                       const unsigned char *secret, unsigned int secret_len,
+                       const unsigned char *context, unsigned int context_len,
+                       unsigned char *outbuf, unsigned int *outbuf_len)
+{
+    const char *info_str = "UFTP5 ";
+    unsigned int info_str_len = (unsigned)strlen(info_str); 
+    unsigned int label_len = (unsigned)strlen(label);
+    unsigned int info_len = info_str_len + label_len + context_len;
+    unsigned char *info = safe_malloc(info_len);
+    
+    memcpy(info, info_str, info_str_len);
+    memcpy(info + info_str_len, label, label_len);
+    memcpy(info + info_str_len + label_len, context, context_len);
+    HKDF_Expand(hashtype, bytes, secret, secret_len, info, info_len, 
+                outbuf, outbuf_len);
+    free(info);
+}
+
+/**
+ * Creates the handshake keys for the client and server
+ */
+void calculate_hs_keys(int hashtype, uint8_t *premaster, int premaster_len,
+                       uint8_t *client_context1, 
+                       unsigned int client_context1_len, 
+                       unsigned int key_len, unsigned int iv_len, 
+                       uint8_t *server_hs_key, uint8_t *server_hs_iv, 
+                       uint8_t *client_hs_key, uint8_t *client_hs_iv)
+{
+    uint8_t zeros[HASH_LEN] = { 0 };
+    uint8_t context_hash[HASH_LEN];
+    uint8_t hs_secret[2*HASH_LEN];
+    uint8_t server_hs_secret[2*HASH_LEN];
+    uint8_t client_hs_secret[2*HASH_LEN];
+    uint8_t tmp_out[2*HASH_LEN];
+    unsigned int hash_len, hs_secret_len, out_len;
+
+    hash_len = get_hash_len(hashtype);
+    hash(hashtype, client_context1, client_context1_len, context_hash,&out_len);
+
+    HKDF_Extract(hashtype, zeros, hash_len, premaster, premaster_len,
+                 hs_secret, &hs_secret_len);
+    HKDF_Expand_Label(hashtype, hash_len, "s hs traffic", hs_secret,
+            hs_secret_len, context_hash, hash_len, server_hs_secret, &out_len);
+    HKDF_Expand_Label(hashtype, hash_len, "c hs traffic", hs_secret,
+            hs_secret_len, context_hash, hash_len, client_hs_secret, &out_len);
+
+    HKDF_Expand_Label(hashtype, key_len, "key", server_hs_secret, hash_len,
+                      (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(server_hs_key, tmp_out, key_len);
+    HKDF_Expand_Label(hashtype, iv_len, "iv", server_hs_secret, hash_len,
+                      (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(server_hs_iv, tmp_out, iv_len);
+    HKDF_Expand_Label(hashtype, key_len, "key", client_hs_secret, hash_len,
+                      (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(client_hs_key, tmp_out, key_len);
+    HKDF_Expand_Label(hashtype, iv_len, "iv", client_hs_secret, hash_len,
+                      (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(client_hs_iv, tmp_out, iv_len);
+}
+
+/**
+ * Creates the application keys for the server
+ */
+void calculate_server_app_keys(int hashtype, uint8_t *groupmaster, 
+                               int groupmaster_len, uint8_t *server_context, 
+                               unsigned int server_context_len,
+                               unsigned int key_len, unsigned int iv_len, 
+                               uint8_t *server_app_key, uint8_t *server_app_iv)
+{
+    uint8_t zeros[HASH_LEN] = { 0 };
+    uint8_t context_hash[HASH_LEN];
+    uint8_t app_secret[HASH_LEN];
+    uint8_t server_app_secret[2*HASH_LEN];
+    uint8_t tmp_out[2*HASH_LEN];
+    unsigned int hash_len, app_secret_len, out_len;
+
+    hash_len = get_hash_len(hashtype);
+    hash(hashtype, server_context, server_context_len, context_hash, &out_len);
+
+    HKDF_Extract(hashtype, zeros, hash_len, groupmaster, groupmaster_len,
+                 app_secret, &app_secret_len);
+    HKDF_Expand_Label(hashtype, hash_len, "s app traffic", app_secret,
+            app_secret_len, context_hash, hash_len, server_app_secret,&out_len);
+
+    HKDF_Expand_Label(hashtype, key_len, "key", server_app_secret, hash_len,
+                      (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(server_app_key, tmp_out, key_len);
+    HKDF_Expand_Label(hashtype, iv_len, "iv", server_app_secret, hash_len,
+                      (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(server_app_iv, tmp_out, iv_len);
+}
+
+/**
+ * Creates the application keys for a client
+ * Also calculates the finished hash for the client
+ */
+void calculate_client_app_keys(int hashtype, uint8_t *groupmaster,
+                               int groupmaster_len, uint8_t *client_context2,
+                               unsigned int client_context2_len,
+                               unsigned int key_len, unsigned int iv_len, 
+                               uint8_t *client_app_key, uint8_t *client_app_iv,
+                               uint8_t *finished_key, uint8_t *verify_data)
+{
+    uint8_t zeros[HASH_LEN] = { 0 };
+    uint8_t context_hash[HASH_LEN];
+    uint8_t app_secret[HASH_LEN];
+    uint8_t client_app_secret[2*HASH_LEN];
+    uint8_t tmp_out[2*HASH_LEN];
+    uint8_t *verify_context;
+    unsigned int hash_len, app_secret_len, verify_context_len, out_len;
+
+    hash_len = get_hash_len(hashtype);
+    hash(hashtype, client_context2, client_context2_len, context_hash,&out_len);
+
+    HKDF_Extract(hashtype, zeros, hash_len, groupmaster, groupmaster_len,
+                 app_secret, &app_secret_len);
+    HKDF_Expand_Label(hashtype, hash_len, "c app traffic", app_secret,
+            app_secret_len, context_hash, hash_len, client_app_secret,&out_len);
+
+    HKDF_Expand_Label(hashtype, key_len, "key", client_app_secret, hash_len,
+                      (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(client_app_key, tmp_out, key_len);
+    HKDF_Expand_Label(hashtype, iv_len, "iv", client_app_secret, hash_len,
+                      (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(client_app_iv, tmp_out, iv_len);
+    HKDF_Expand_Label(hashtype, hash_len, "finished", client_app_secret,
+            hash_len, (const unsigned char *)"", 0, tmp_out, &out_len);
+    memcpy(finished_key, tmp_out, hash_len);
+
+    verify_context_len = 0;
+    verify_context = safe_malloc(client_context2_len + groupmaster_len);
+    memcpy(verify_context + verify_context_len, client_context2, 
+            client_context2_len);
+    verify_context_len += client_context2_len;
+    memcpy(verify_context + verify_context_len, groupmaster, groupmaster_len);
+    verify_context_len += groupmaster_len;
+
+    create_hmac(hashtype, finished_key, hash_len, verify_context,
+                verify_context_len, verify_data, &out_len);
+    free(verify_context);
 }
 
 /**
@@ -1981,7 +2076,7 @@ uint64_t free_space(const char *file)
         free(filename);
         return 0x7FFFFFFFFFFFFFFFULL;
     } else {
-        log3(0, 0, 0, "Free space: %s", printll(bytes_free.QuadPart));
+        log3(0, 0, 0, "Free space: " F_i64, bytes_free.QuadPart);
         free(dirname);
         free(filename);
         return bytes_free.QuadPart;
@@ -1993,8 +2088,8 @@ uint64_t free_space(const char *file)
         syserror(0, 0, 0, "statvfs failed");
         return 0x7FFFFFFFFFFFFFFFULL;
     } else {
-        log3(0, 0, 0, "Free space: %s",
-                      printll((uint64_t)buf.f_bsize * buf.f_bavail));
+        log3(0, 0, 0, "Free space: " F_i64,
+                      (uint64_t)buf.f_bsize * buf.f_bavail);
         return (uint64_t)buf.f_bsize * buf.f_bavail;
     }
 #endif
@@ -2028,7 +2123,7 @@ int valid_priority(int priority)
  */
 uint32_t rand32()
 {
-    return((rand() & 0x7FFF) << 17) | ((rand() & 0x7FFF) << 2) | (rand() & 0x2);
+    return((rand() & 0x7FFF) << 17) | ((rand() & 0x7FFF) << 2) | (rand() & 0x3);
 }
 
 /**
